# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import click

from xcube.cli.common import (
    cli_option_quiet,
    cli_option_verbosity,
    configure_cli_output,
    parse_cli_kwargs,
)

DEFAULT_OUTPUT_PATH = "out.zarr"


# noinspection PyShadowingBuiltins
@click.command(name="chunk")
@click.argument("cube")
@click.option(
    "--output",
    "-o",
    metavar="OUTPUT",
    default=DEFAULT_OUTPUT_PATH,
    help=f"Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}",
)
@click.option(
    "--format",
    "-f",
    metavar="FORMAT",
    type=click.Choice(["zarr", "netcdf"]),
    help="Format of the output. If not given, guessed from OUTPUT.",
)
@click.option(
    "--params",
    "-p",
    metavar="PARAMS",
    help="Parameters specific for the output format."
    " Comma-separated list of <key>=<value> pairs.",
)
@click.option(
    "--chunks",
    "-C",
    metavar="CHUNKS",
    nargs=1,
    default=None,
    help="Chunk sizes for each dimension."
    " Comma-separated list of <dim>=<size> pairs,"
    ' e.g. "time=1,lat=270,lon=270"',
)
@cli_option_quiet
@cli_option_verbosity
def chunk(
    cube, output, format=None, params=None, chunks=None, quiet=None, verbosity=None
):
    """
    (Re-)chunk xcube dataset.
    Changes the external chunking of all variables of CUBE according to CHUNKS and writes
    the result to OUTPUT.

    Note: There is a possibly more efficient way to (re-)chunk datasets through the
    dedicated tool "rechunker", see https://rechunker.readthedocs.io.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    chunk_sizes = None
    if chunks:
        chunk_sizes = parse_cli_kwargs(chunks, metavar="CHUNKS")
        for k, v in chunk_sizes.items():
            if not isinstance(v, int) or v <= 0:
                raise click.ClickException(
                    "Invalid value for CHUNKS, "
                    f"chunk sizes must be positive integers: {chunks}"
                )

    write_kwargs = dict()
    if params:
        write_kwargs = parse_cli_kwargs(params, metavar="PARAMS")

    from xcube.core.chunk import chunk_dataset
    from xcube.core.dsio import guess_dataset_format, open_dataset, write_dataset

    format_name = format if format else guess_dataset_format(output)

    with open_dataset(input_path=cube) as ds:
        if chunk_sizes:
            for k in chunk_sizes:
                if k not in ds.sizes:
                    raise click.ClickException(
                        "Invalid value for CHUNKS, "
                        f"{k!r} is not the name of any dimension: {chunks}"
                    )

        chunked_dataset = chunk_dataset(
            ds, chunk_sizes=chunk_sizes, format_name=format_name
        )
        write_dataset(
            chunked_dataset, output_path=output, format_name=format_name, **write_kwargs
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Dict

import pandas as pd
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.version import version

from ..config import CubeConfig
from .transformer import CubeTransformer, TransformedCube


class CubeMetadataAdjuster(CubeTransformer):
    """Adjust a cube's metadata."""

    def transform_cube(
        self, cube: xr.Dataset, gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        history = cube.attrs.get("history")
        if isinstance(history, str):
            history = [history]
        elif isinstance(history, (list, tuple)):
            history = list(history)
        else:
            history = []
        history.append(
            dict(
                program=f"xcube gen2, version {version}",
                cube_config=cube_config.to_dict(),
            )
        )
        cube = cube.assign_attrs(
            Conventions="CF-1.7",
            history=history,
            date_created=pd.Timestamp.now().isoformat(),
            # TODO: adjust temporal metadata too
            **get_geospatial_attrs(gm),
        )
        if cube_config.metadata:
            self._check_for_self_destruction(cube_config.metadata)
            cube.attrs.update(cube_config.metadata)
        if cube_config.variable_metadata:
            for var_name, metadata in cube_config.variable_metadata.items():
                if var_name in cube.variables and metadata:
                    cube[var_name].attrs.update(metadata)
        return cube, gm, cube_config

    @staticmethod
    def _check_for_self_destruction(metadata: dict[str, Any]):
        key = "inverse_fine_structure_constant"
        value = 137
        if key in metadata and metadata[key] != value:
            # Note, this is an easter egg that causes
            # an intended internal error for testing
            raise ValueError(f"{key} must be {value} or running in wrong universe")


def get_geospatial_attrs(gm: GridMapping) -> dict[str, Any]:
    return dict(gm.to_dataset_attrs())



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.constants import (
    EXTENSION_POINT_CLI_COMMANDS,
    EXTENSION_POINT_DATA_OPENERS,
    EXTENSION_POINT_DATA_STORES,
    EXTENSION_POINT_DATA_WRITERS,
    EXTENSION_POINT_DATASET_IOS,
    EXTENSION_POINT_INPUT_PROCESSORS,
    EXTENSION_POINT_SERVER_APIS,
    EXTENSION_POINT_SERVER_FRAMEWORKS,
    FORMAT_NAME_CSV,
    FORMAT_NAME_MEM,
    FORMAT_NAME_NETCDF4,
    FORMAT_NAME_ZARR,
)
from xcube.util import extension


def init_plugin(ext_registry: extension.ExtensionRegistry):
    """Register xcube's standard extensions."""
    _register_input_processors(ext_registry)
    _register_dataset_ios(ext_registry)
    _register_cli_commands(ext_registry)
    _register_data_stores(ext_registry)
    _register_data_accessors(ext_registry)
    _register_server_apis(ext_registry)
    _register_server_frameworks(ext_registry)


def _register_input_processors(ext_registry: extension.ExtensionRegistry):
    """Register xcube's standard input processors used by "xcube gen" and gen_cube()."""
    ext_registry.add_extension(
        loader=extension.import_component("xcube.core.gen.iproc:DefaultInputProcessor"),
        point=EXTENSION_POINT_INPUT_PROCESSORS,
        name="default",
        description="Single-scene NetCDF/CF inputs in xcube standard format",
    )


def _register_dataset_ios(ext_registry: extension.ExtensionRegistry):
    """Register xcube's standard dataset I/O components
    used by various CLI and API functions."""
    ext_registry.add_extension(
        loader=extension.import_component("xcube.core.dsio:ZarrDatasetIO", call=True),
        point=EXTENSION_POINT_DATASET_IOS,
        name=FORMAT_NAME_ZARR,
        description="Zarr file format (http://zarr.readthedocs.io)",
        ext="zarr",
        modes={"r", "w", "a"},
    )
    ext_registry.add_extension(
        loader=extension.import_component(
            "xcube.core.dsio:Netcdf4DatasetIO", call=True
        ),
        point=EXTENSION_POINT_DATASET_IOS,
        name=FORMAT_NAME_NETCDF4,
        description="NetCDF-4 file format",
        ext="nc",
        modes={"r", "w", "a"},
    )
    ext_registry.add_extension(
        loader=extension.import_component("xcube.core.dsio:CsvDatasetIO", call=True),
        point=EXTENSION_POINT_DATASET_IOS,
        name=FORMAT_NAME_CSV,
        description="CSV file format",
        ext="csv",
        modes={"r", "w"},
    )
    ext_registry.add_extension(
        loader=extension.import_component("xcube.core.dsio:MemDatasetIO", call=True),
        point=EXTENSION_POINT_DATASET_IOS,
        name=FORMAT_NAME_MEM,
        description="In-memory dataset I/O",
        ext="mem",
        modes={"r", "w", "a"},
    )


_FS_STORAGE_ITEMS = (
    ("abfs", "Azure blob compatible object storage"),
    ("file", "local filesystem"),
    ("ftp", "FTP filesystem"),
    ("https", "HTTPS filesystem"),
    ("memory", "in-memory filesystem"),
    ("reference", "reference filesystem"),
    ("s3", "AWS S3 compatible object storage"),
)

_FS_DATA_ACCESSOR_ITEMS = (
    ("dataset", "netcdf", "xarray.Dataset in NetCDF format"),
    ("dataset", "zarr", "xarray.Dataset in Zarr format"),
    ("dataset", "levels", "xarray.Dataset in leveled Zarr format"),
    (
        "mldataset",
        "levels",
        "xcube.core.mldataset.MultiLevelDataset in leveled Zarr format",
    ),
    ("dataset", "geotiff", "xarray.Dataset in GeoTIFF or COG format"),
    (
        "mldataset",
        "geotiff",
        "xcube.core.mldataset.MultiLevelDataset in GeoTIFF or COG format",
    ),
    ("geodataframe", "shapefile", "gpd.GeoDataFrame in ESRI Shapefile format"),
    ("geodataframe", "geojson", "gpd.GeoDataFrame in GeoJSON format"),
)

_FS_DATA_OPENER_ITEMS = _FS_DATA_ACCESSOR_ITEMS
_FS_DATA_WRITER_ITEMS = _FS_DATA_ACCESSOR_ITEMS


def _register_data_stores(ext_registry: extension.ExtensionRegistry):
    """Register xcube's standard data stores."""
    fs_ds_cls_factory = "xcube.core.store.fs.registry:get_fs_data_store_class"
    for storage_id, storage_description in _FS_STORAGE_ITEMS:
        fs_ds_cls_loader = extension.import_component(
            fs_ds_cls_factory, call_args=[storage_id]
        )
        ext_registry.add_extension(
            point=EXTENSION_POINT_DATA_STORES,
            loader=fs_ds_cls_loader,
            name=storage_id,
            description=f"Data store that uses a {storage_description}",
        )

    ref_ds_cls = "xcube.core.store.ref.store:ReferenceDataStore"
    ref_ds_cls_loader = extension.import_component(ref_ds_cls)
    ext_registry.add_extension(
        point=EXTENSION_POINT_DATA_STORES,
        loader=ref_ds_cls_loader,
        name="reference",
        description="Data store that uses Kerchunk references",
    )


def _register_data_accessors(ext_registry: extension.ExtensionRegistry):
    """Register xcube's standard data accessors."""
    factory = "xcube.core.store.fs.registry:get_fs_data_accessor_class"

    # noinspection PyShadowingNames
    def _add_fs_data_accessor_ext(
        point: str, ext_type: str, protocol: str, data_type: str, format_id: str
    ):
        factory_args = (protocol, data_type, format_id)
        loader = extension.import_component(factory, call_args=factory_args)
        ext_registry.add_extension(
            point=point,
            loader=loader,
            name=f"{data_type}:{format_id}:{protocol}",
            description=f"Data {ext_type} for"
            f" a {data_accessor_description}"
            f" in {storage_description}",
        )

    for protocol, storage_description in _FS_STORAGE_ITEMS:
        for data_type, format_id, data_accessor_description in _FS_DATA_OPENER_ITEMS:
            _add_fs_data_accessor_ext(
                EXTENSION_POINT_DATA_OPENERS, "opener", protocol, data_type, format_id
            )
        for data_type, format_id, data_accessor_description in _FS_DATA_WRITER_ITEMS:
            _add_fs_data_accessor_ext(
                EXTENSION_POINT_DATA_WRITERS, "writer", protocol, data_type, format_id
            )


def _register_cli_commands(ext_registry: extension.ExtensionRegistry):
    """Register xcube's standard CLI commands."""

    cli_command_names = [
        "chunk",
        "compute",
        "benchmark",
        "dump",
        "extract",
        "gen",
        "gen2",
        "genpts",
        "grid",
        "level",
        "optimize",
        "patch",
        "prune",
        "rectify",
        "resample",
        "serve",
        "vars2dim",
        "verify",
        "versions",
        # Experimental + Hidden
        "io",
    ]

    for cli_command_name in cli_command_names:
        ext_registry.add_extension(
            loader=extension.import_component(
                f"xcube.cli.{cli_command_name}:{cli_command_name}"
            ),
            point=EXTENSION_POINT_CLI_COMMANDS,
            name=cli_command_name,
        )


def _register_server_apis(ext_registry: extension.ExtensionRegistry):
    """Register xcube's standard server APIs."""
    server_api_names = [
        "meta",
        "auth",
        "compute",
        "places",
        "styles",
        "datasets",
        "tiles",
        "timeseries",
        "statistics",
        "volumes",
        "expressions",
        "ows.coverages",
        "ows.stac",
        "ows.wmts",
        "s3",
        "viewer",
    ]
    for api_name in server_api_names:
        ext_registry.add_extension(
            loader=extension.import_component(f"xcube.webapi.{api_name}:api"),
            point=EXTENSION_POINT_SERVER_APIS,
            name=api_name,
        )


def _register_server_frameworks(ext_registry: extension.ExtensionRegistry):
    server_framework_names = [
        "tornado",
        "flask",
    ]
    for framework_name in server_framework_names:
        ext_registry.add_extension(
            loader=extension.import_component(
                f"xcube.server.webservers.{framework_name}"
                f":{framework_name.capitalize()}Framework",
            ),
            point=EXTENSION_POINT_SERVER_FRAMEWORKS,
            name=framework_name,
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

DEFAULT_OUTPUT_PATH = "out.zarr"
DEFAULT_OUTPUT_RESAMPLING = "Nearest"
DEFAULT_OUTPUT_SIZE = [512, 512]



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Dict, Optional, Union

import shapely.geometry
import shapely.wkt
from shapely.lib import ShapelyError

from xcube.constants import LOG
from xcube.core.geom import get_box_split_bounds_geometry
from xcube.util.perf import measure_time

from ...server.api import ApiError
from ...util.assertions import assert_instance
from .context import ALL_PLACES, PlacesContext

GeoJsonFeatureCollection = dict
GeoJsonFeature = dict

GeometryLike = Union[str, dict, shapely.geometry.base.BaseGeometry]


def find_places(
    ctx: PlacesContext,
    place_group_id: str,
    base_url: str,
    query_geometry: Optional[GeometryLike] = None,
    query_expr: Optional[Any] = None,
    comb_op: str = "and",
) -> GeoJsonFeatureCollection:
    try:
        if isinstance(query_geometry, str):
            try:
                query_geometry = get_box_split_bounds_geometry(
                    *[float(s) for s in query_geometry.split(",")]
                )
            except (TypeError, ValueError):
                query_geometry = shapely.wkt.loads(query_geometry)
        elif isinstance(query_geometry, dict):
            if query_geometry["type"] == "FeatureCollection":
                query_geometry = shapely.geometry.shape(
                    query_geometry["features"][0]["geometry"]
                )
            elif query_geometry["type"] == "Feature":
                query_geometry = shapely.geometry.shape(query_geometry["geometry"])
            else:
                query_geometry = shapely.geometry.shape(query_geometry)
        elif query_geometry is not None:
            assert_instance(
                query_geometry,
                shapely.geometry.base.BaseGeometry,
                name="query_geometry",
            )
    except (ShapelyError, TypeError, IndexError, ValueError, KeyError) as e:
        raise ApiError.BadRequest(
            "Received invalid geometry bbox, geometry WKT, or GeoJSON object"
        ) from e

    with measure_time() as cm:
        features = _find_places(
            ctx, place_group_id, base_url, query_geometry, query_expr, comb_op
        )
    LOG.info(f"{len(features)} places found within {cm.duration} seconds")
    return features


def _find_places(
    ctx: PlacesContext,
    place_group_id: str,
    base_url: str,
    query_geometry: Optional[shapely.geometry.base.BaseGeometry] = None,
    query_expr: Optional[Any] = None,
    comb_op: str = "and",
) -> GeoJsonFeatureCollection:
    if comb_op is not None and comb_op != "and":
        raise ApiError.NotImplemented("comb_op not yet supported")

    if place_group_id == ALL_PLACES:
        place_groups = ctx.get_global_place_groups(base_url, load_features=True)
        features = []
        for place_group in place_groups:
            features.extend(place_group["features"])
        feature_collection = dict(type="FeatureCollection", features=features)
    else:
        feature_collection = ctx.get_global_place_group(
            place_group_id, base_url, load_features=True
        )
        feature_collection = dict(
            type="FeatureCollection", features=feature_collection["features"]
        )

    if query_geometry is None:
        if query_expr is None:
            return feature_collection
        else:
            raise ApiError.NotImplemented("query_expr not yet supported")
    else:
        matching_places = []
        if query_expr is None:
            for feature in feature_collection["features"]:
                geometry = shapely.geometry.shape(feature["geometry"])
                if geometry.intersects(query_geometry):
                    matching_places.append(feature)
        else:
            raise NotImplementedError()
        return dict(type="FeatureCollection", features=matching_places)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Sequence

from xcube.core.store import (
    DATASET_TYPE,
    DatasetDescriptor,
    DataStoreError,
    DataStorePool,
    get_data_store_instance,
    new_data_opener,
)
from xcube.util.assertions import assert_instance, assert_true
from xcube.util.progress import observe_progress

from ..config import InputConfig
from ..error import CubeGeneratorError


class DatasetsDescriber:
    def __init__(
        self, input_configs: Sequence[InputConfig], store_pool: DataStorePool = None
    ):
        assert_true(len(input_configs) > 0, "At least one input must be given")
        if store_pool is not None:
            assert_instance(store_pool, DataStorePool, "store_pool")
        self._input_configs = input_configs
        self._store_pool = store_pool

    def describe_datasets(self) -> Sequence[DatasetDescriptor]:
        descriptors = []
        with observe_progress(
            "Fetching dataset information", len(self._input_configs)
        ) as progress:
            for input_config in self._input_configs:
                descriptors.append(self._describe_dataset(input_config))
                progress.worked(1)
        return descriptors

    def _describe_dataset(self, input_config: InputConfig) -> DatasetDescriptor:
        opener_id = input_config.opener_id
        store_params = input_config.store_params or {}
        if input_config.store_id:
            store_instance = get_data_store_instance(
                input_config.store_id,
                store_params=store_params,
                store_pool=self._store_pool,
            )
            opener = store_instance.store
        else:
            opener = new_data_opener(opener_id)
        try:
            descriptor = opener.describe_data(
                input_config.data_id, data_type=DATASET_TYPE
            )
        except DataStoreError as dse:
            raise CubeGeneratorError(f"{dse}", status_code=400) from dse
        if not isinstance(descriptor, DatasetDescriptor):
            raise RuntimeError(
                f"internal error: data store "
                f'"{input_config.store_id}": '
                f"expected DatasetDescriptor but got "
                f"a {type(descriptor)}"
            )
        return descriptor



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import ABC
from typing import Optional, Tuple

import fsspec
import rasterio
import rioxarray
import s3fs
import xarray as xr
import zarr
from rasterio.session import AWSSession

# Note, we need the following reference to register the
# xarray property accessor
# noinspection PyUnresolvedReferences
from xcube.core.zarrstore import LoggingZarrStore, ZarrStoreHolder
from xcube.util.assertions import assert_instance, assert_true
from xcube.util.fspath import is_https_fs, is_local_fs
from xcube.util.jsonencoder import to_json_value
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonBooleanSchema,
    JsonIntegerSchema,
    JsonNumberSchema,
    JsonObjectSchema,
    JsonStringSchema,
)
from xcube.util.temp import new_temp_file

from ...datatype import DATASET_TYPE, DataType
from ...error import DataStoreError
from ..accessor import FsDataAccessor

ZARR_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        log_access=JsonBooleanSchema(default=False),
        cache_size=JsonIntegerSchema(
            minimum=0,
        ),
        group=JsonStringSchema(
            description="Group path. (a.k.a. path in zarr terminology.).",
            min_length=1,
        ),
        engine=JsonStringSchema(
            description="xarray backend name.",
            min_length=1,
        ),
        chunks=JsonObjectSchema(
            description="Optional chunk sizes along each dimension."
            ' Chunk size values may be None, "auto"'
            " or an integer value.",
            examples=[
                {"time": None, "lat": "auto", "lon": 90},
                {"time": 1, "y": 512, "x": 512},
            ],
            additional_properties=True,
        ),
        decode_cf=JsonBooleanSchema(
            description="Whether to decode these variables,"
            " assuming they were saved according to"
            " CF conventions.",
            default=True,
        ),
        mask_and_scale=JsonBooleanSchema(
            description="If True, replace array values equal"
            ' to attribute "_FillValue" with NaN. '
            ' Use "scale_factor" and "add_offset"'
            " attributes to compute actual values.",
            default=True,
        ),
        decode_times=JsonBooleanSchema(
            description="If True, decode times encoded in the"
            " standard NetCDF datetime format "
            "into datetime objects. Otherwise,"
            " leave them encoded as numbers.",
            default=True,
        ),
        decode_coords=JsonBooleanSchema(
            description='If True, decode the "coordinates"'
            " attribute to identify coordinates in "
            "the resulting dataset.",
            default=True,
        ),
        drop_variables=JsonArraySchema(
            items=JsonStringSchema(min_length=1),
        ),
        consolidated=JsonBooleanSchema(
            description="Whether to open the store using"
            " Zarr's consolidated metadata "
            "capability. Only works for stores that"
            " have already been consolidated.",
            default=False,
        ),
    ),
    required=[],
    # additional_properties=True because we want to allow passing arbitrary
    # parameters to xarray.open_dataset()
    additional_properties=True,
)

ZARR_WRITE_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        group=JsonStringSchema(
            description="Group path. (a.k.a. path in zarr terminology.).",
            min_length=1,
        ),
        encoding=JsonObjectSchema(
            description="Nested dictionary with variable"
            " names as keys and dictionaries of"
            " variable specific encodings as values.",
            examples=[{"my_variable": {"dtype": "int16", "scale_factor": 0.1}}],
            additional_properties=True,
        ),
        consolidated=JsonBooleanSchema(
            description="If True (the default), consolidate all metadata"
            ' files ("**/.zarray", "**/.zattrs")'
            ' into a single top-level file ".zmetadata"',
            default=True,
        ),
        append_dim=JsonStringSchema(
            description="If set, the dimension on which the data will be appended.",
            min_length=1,
        ),
        replace=JsonBooleanSchema(
            description="If set, an existing dataset will be replaced without warning.",
        ),
    ),
    # additional_properties=True because we want to allow passing arbitrary
    # parameters to xarray.to_zarr()
    additional_properties=True,
)


class DatasetFsDataAccessor(FsDataAccessor, ABC):
    """Opener/writer extension name: 'dataset:<format>:<protocol>'."""

    @classmethod
    def get_data_type(cls) -> DataType:
        return DATASET_TYPE


class DatasetZarrFsDataAccessor(DatasetFsDataAccessor):
    """Opener/writer extension name: 'dataset:zarr:<protocol>'."""

    @classmethod
    def get_format_id(cls) -> str:
        return "zarr"

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_open_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(ZARR_OPEN_DATA_PARAMS_SCHEMA)

    def open_data(self, data_id: str, **open_params) -> xr.Dataset:
        assert_instance(data_id, str, name="data_id")
        fs, root, open_params = self.load_fs(open_params)
        engine = open_params.pop("engine", "zarr")
        cache_size = open_params.pop("cache_size", None)
        log_access = open_params.pop("log_access", None)
        consolidated = open_params.pop(
            "consolidated", fs.exists(f"{data_id}/.zmetadata")
        )
        zarr_store = fs.get_mapper(data_id)
        if isinstance(cache_size, int) and cache_size > 0:
            zarr_store = zarr.LRUStoreCache(zarr_store, max_size=cache_size)
        if log_access:
            zarr_store = LoggingZarrStore(zarr_store, name=f"zarr_store({data_id!r})")
        # TODO: test whether we really need to distinguish here as we know
        #   we are opening a zarr dataset, even without another backend.
        if engine == "zarr":
            try:
                dataset = xr.open_zarr(
                    zarr_store, consolidated=consolidated, **open_params
                )
            except ValueError as e:
                raise DataStoreError(f"Failed to open dataset {data_id!r}: {e}") from e
        else:
            try:
                dataset = xr.open_dataset(zarr_store, engine=engine, **open_params)
            except ValueError as e:
                raise DataStoreError(
                    f"Failed to open dataset {data_id!r} using engine {engine!r}: {e}"
                ) from e

        dataset.zarr_store.set(zarr_store)

        return dataset

    # noinspection PyMethodMayBeStatic
    def get_write_data_params_schema(self) -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(ZARR_WRITE_DATA_PARAMS_SCHEMA)

    def write_data(
        self, data: xr.Dataset, data_id: str, replace=False, **write_params
    ) -> str:
        assert_instance(data, xr.Dataset, name="data")
        assert_instance(data_id, str, name="data_id")
        fs, root, write_params = self.load_fs(write_params)
        zarr_store = fs.get_mapper(data_id, create=True)
        log_access = write_params.pop("log_access", None)
        if log_access:
            zarr_store = LoggingZarrStore(zarr_store, name=f"zarr_store({data_id!r})")
        consolidated = write_params.pop("consolidated", True)
        try:
            data.to_zarr(
                zarr_store,
                mode="w" if replace else None,
                consolidated=consolidated,
                **write_params,
            )
        except ValueError as e:
            raise DataStoreError(f"Failed to write dataset {data_id!r}: {e}") from e
        return data_id

    def delete_data(self, data_id: str, **delete_params):
        fs, root, delete_params = self.load_fs(delete_params)
        delete_params.pop("recursive", None)
        fs.delete(data_id, recursive=True, **delete_params)


NETCDF_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        # TODO: add more from xr.open_dataset()
    ),
    additional_properties=True,
)

NETCDF_WRITE_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        # TODO: add more from ds.to_netcdf()
    ),
    additional_properties=True,
)


class DatasetNetcdfFsDataAccessor(DatasetFsDataAccessor):
    """Opener/writer extension name: 'dataset:netcdf:<protocol>'."""

    @classmethod
    def get_format_id(cls) -> str:
        return "netcdf"

    def get_open_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(NETCDF_OPEN_DATA_PARAMS_SCHEMA)

    def open_data(self, data_id: str, **open_params) -> xr.Dataset:
        assert_instance(data_id, str, name="data_id")
        fs, root, open_params = self.load_fs(open_params)

        # This doesn't yet work as expected with fsspec and netcdf:
        # engine = open_params.pop('engine', 'scipy')
        # with fs.open(data_id, 'rb') as file:
        #     return xr.open_dataset(file, engine=engine, **open_params)

        if is_local_fs(fs):
            file_path = data_id
        elif is_https_fs(fs):
            file_path = f"{fs.protocol}://{data_id}#mode=bytes"
        else:
            _, file_path = new_temp_file(suffix=".nc")
            fs.get_file(data_id, file_path)
        engine = open_params.pop("engine", "netcdf4")
        return xr.open_dataset(file_path, engine=engine, **open_params)

    def get_write_data_params_schema(self) -> JsonObjectSchema:
        return self.add_storage_options_to_params_schema(
            NETCDF_WRITE_DATA_PARAMS_SCHEMA
        )

    def write_data(
        self, data: xr.Dataset, data_id: str, replace=False, **write_params
    ) -> str:
        assert_instance(data, xr.Dataset, name="data")
        assert_instance(data_id, str, name="data_id")
        fs, root, write_params = self.load_fs(write_params)
        if not replace and fs.exists(data_id):
            raise DataStoreError(f"Data resource {data_id} already exists")

        # This doesn't yet work as expected with fsspec and netcdf:
        # engine = write_params.pop('engine', 'scipy')
        # with fs.open(data_id, 'wb') as file:
        #     data.to_netcdf(file, engine=engine, **write_params)

        is_local = is_local_fs(fs)
        if is_local:
            file_path = data_id
        else:
            _, file_path = new_temp_file(suffix=".nc")
        engine = write_params.pop("engine", "netcdf4")
        data.to_netcdf(file_path, engine=engine, **write_params)
        if not is_local:
            fs.put_file(file_path, data_id)
        return data_id


GEOTIFF_OPEN_DATA_PARAMS_SCHEMA = JsonObjectSchema(
    properties=dict(
        tile_size=JsonArraySchema(
            items=(
                JsonNumberSchema(minimum=256, default=512),
                JsonNumberSchema(minimum=256, default=512),
            ),
            default=[512, 512],
        ),
        overview_level=JsonIntegerSchema(
            default=None,
            nullable=True,
            description="GeoTIFF overview level. 0 is the first overview.",
        ),
    ),
    additional_properties=False,
)


class DatasetGeoTiffFsDataAccessor(DatasetFsDataAccessor):
    """
    Opener/writer extension name: 'dataset:geotiff:<protocol>'.
    """

    @classmethod
    def get_format_id(cls) -> str:
        return "geotiff"

    def get_open_data_params_schema(self, data_id: str = None) -> JsonObjectSchema:
        return GEOTIFF_OPEN_DATA_PARAMS_SCHEMA

    def open_data(self, data_id: str, **open_params) -> xr.Dataset:
        assert_instance(data_id, str, name="data_id")
        fs, root, open_params = self.load_fs(open_params)

        if isinstance(fs.protocol, str):
            protocol = fs.protocol
        else:
            protocol = fs.protocol[0]
        if root is not None:
            file_path = protocol + "://" + root + "/" + data_id
        else:
            file_path = protocol + "://" + data_id
        tile_size = open_params.get("tile_size", (512, 512))
        overview_level = open_params.get("overview_level", None)
        return self.open_dataset(
            fs, file_path, tile_size, overview_level=overview_level
        )

    @classmethod
    def create_env_session(cls, fs):
        if isinstance(fs, s3fs.S3FileSystem):
            aws_unsigned = bool(fs.anon)
            aws_session = AWSSession(
                aws_unsigned=aws_unsigned,
                aws_secret_access_key=fs.secret,
                aws_access_key_id=fs.key,
                aws_session_token=fs.token,
                region_name=fs.client_kwargs.get("region_name", "eu-central-1"),
            )
            return rasterio.env.Env(
                session=aws_session, aws_no_sign_request=aws_unsigned
            )
        return rasterio.env.NullContextManager()

    @classmethod
    def open_dataset_with_rioxarray(
        cls, file_path, overview_level, tile_size
    ) -> rioxarray.raster_array:
        return rioxarray.open_rasterio(
            file_path,
            overview_level=overview_level,
            chunks=dict(zip(("x", "y"), tile_size)),
        )

    @classmethod
    def open_dataset(
        cls,
        fs,
        file_path: str,
        tile_size: tuple[int, int],
        overview_level: Optional[int] = None,
    ) -> xr.Dataset:
        """
        A method to open the cog/geotiff dataset using rioxarray,
        returns xarray.Dataset
        @param fs: abstract file system
        @type fs: fsspec.AbstractFileSystem object.
        @param file_path: path of the file
        @type file_path: str
        @param overview_level: the overview level of GeoTIFF, 0 is the first
               overview and None means full resolution.
        @type overview_level: int
        @param tile_size: tile size as tuple.
        @type tile_size: tuple
        """

        if isinstance(fs, fsspec.AbstractFileSystem):
            with cls.create_env_session(fs):
                array = cls.open_dataset_with_rioxarray(
                    file_path, overview_level, tile_size
                )
        else:
            assert_true(fs is None, message="invalid type for fs")
        arrays = {}
        if array.ndim == 3:
            for i in range(array.shape[0]):
                name = f"{array.name or 'band'}_{i + 1}"
                dims = array.dims[-2:]
                coords = {
                    n: v
                    for n, v in array.coords.items()
                    if n in dims or n == "spatial_ref"
                }
                band_data = array.data[i, :, :]
                arrays[name] = xr.DataArray(
                    band_data, coords=coords, dims=dims, attrs=dict(**array.attrs)
                )
        elif array.ndim == 2:
            name = f"{array.name or 'band'}"
            arrays[name] = array
        else:
            raise RuntimeError("number of dimensions must be 2 or 3")

        dataset = xr.Dataset(arrays, attrs=dict(source=file_path))
        # For CRS, rioxarray uses variable "spatial_ref" by default
        if "spatial_ref" in array.coords:
            for data_var in dataset.data_vars.values():
                data_var.attrs["grid_mapping"] = "spatial_ref"

        # rioxarray may return non-JSON-serializable metadata
        # attribute values.
        # We have seen _FillValue of type np.uint8
        cls._sanitize_dataset_attrs(dataset)

        return dataset

    def get_write_data_params_schema(self) -> JsonObjectSchema:
        raise NotImplementedError("Writing of GeoTIFF not yet supported")

    def write_data(
        self, data: xr.Dataset, data_id: str, replace=False, **write_params
    ) -> str:
        raise NotImplementedError("Writing of GeoTIFF not yet supported")

    @classmethod
    def _sanitize_dataset_attrs(cls, dataset):
        dataset.attrs.update(to_json_value(dataset.attrs))
        for var in dataset.variables.values():
            var.attrs.update(to_json_value(var.attrs))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .affine import affine_transform_dataset, resample_ndimage
from .cf import encode_grid_mapping
from .rectify import rectify_dataset
from .spatial import resample_in_space
from .temporal import resample_in_time



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import inspect
from typing import Any, Callable, Union

import numpy as np
import xarray as xr

from .error import VarExprError
from .exprvar import ExprVar
from .names import _GLOBAL_NAMES, get_constants, get_numpy_ufuncs, get_xarray_funcs
from .varexpr import evaluate


class VarExprContext:
    """Allow safe evaluation of expressions in the context of a
    `xarray.Dataset` object.

    Args:
        dataset: The dataset that provides the variables that can
            be accessed in the expressions passed to :meth:`evaluate`.
    """

    def __init__(self, dataset: xr.Dataset):
        self._array_variables = dict(
            {str(k): ExprVar(v) for k, v in dataset.data_vars.items()}
        )
        self._names = dict(_GLOBAL_NAMES)
        self._names.update(self._array_variables)

    def names(self) -> dict[str, Any]:
        return dict(self._names)

    def get_array_variables(self) -> list[str]:
        """Get array functions."""
        return sorted(self._array_variables.keys())

    @classmethod
    def get_array_functions(cls) -> list[str]:
        """Get array functions."""
        # noinspection PyTypeChecker
        return sorted(
            _format_callable(name, fn)
            for name, fn in dict(**get_xarray_funcs(), **get_numpy_ufuncs()).items()
        )

    @classmethod
    def get_other_functions(cls) -> list[str]:
        """Get other functions that cannot be used with arrays."""
        return []

    @classmethod
    def get_array_operators(cls) -> list[str]:
        """Get array operators."""
        return [
            "+X",
            "-X",
            "~X",
            "X + Y",
            "X - Y",
            "X * Y",
            "X / Y",
            "X // Y",
            "X % Y",
            "X ** Y",
            "X << Y",
            "X >> Y",
            "X & Y",
            "X ^ Y",
            "X | Y",
            "X == Y",
            "X != Y",
            "X < Y",
            "X <= Y",
            "X > Y",
            "X >= Y",
        ]

    @classmethod
    def get_other_operators(cls) -> list[str]:
        """Get other operators that cannot be used with arrays."""
        return [
            "X and Y",
            "X or Y",
            "not X",
            "X in Y",
            "X not in Y",
            "X is Y",
            "X is not Y",
        ]

    @classmethod
    def get_constants(cls) -> list[str]:
        """Get constants."""
        return sorted(get_constants().keys())

    def evaluate(self, var_expr: str) -> xr.DataArray:
        """Evaluate given Python expression *var_expr* in the context of an
        `xarray.Dataset` object.

        The expression *var_expr* may reference the following names:

        * the dataset's data variables;
        * the dataset's coordinate variables;
        * the numpy constants `e`, `pi`, `nan`, `inf`;
        * all numpy ufuncs (https://numpy.org/doc/stable/reference/ufuncs.html);
        * the `where` function (https://docs.xarray.dev/en/stable/generated/xarray.where.html);
        * the Python built-in functions `min`, `max`, `round`, `floor`, `ceil`,
          `bool`, `int`, `float`, `complex`, `str`, `tuple`, `set`, `list`, `dict`.

        In general, all Python numerical and logical operators such as
        `not`, `and`, `or` are supported.
        However, for dataset variables the following subset of operators apply:

        * binary comparison operators: `==`, `!=`, `<`, `<=`, `>`, `>=`;
        * binary arithmetic operators: `+`, `-`, `*`, `/`, `//`, `%`,
            `**`, `<<`, `>>`, `&`, `^`, `|`;
        * unary operators: `+`, `-`, `~`.

        Args:
            var_expr: Expression to be evaluated.

        Returns:
            A newly computed variable of type `xarray.DataArray`.
        """
        try:
            result = evaluate(var_expr, self._names)
        except BaseException as e:
            # Do not report the name 'ExprVar'
            raise VarExprError(f"{e}".replace("ExprVar", "DataArray")) from e

        if not isinstance(result, ExprVar):
            # We do not mention 'ExprVar' by intention
            raise VarExprError(
                f"result must be a 'DataArray' object,"
                f" but got type {result.__class__.__name__!r}"
            )

        result = result.__dict__.get("_ExprVar__da")
        if not isinstance(result, xr.DataArray):
            # noinspection PyUnresolvedReferences
            raise RuntimeError(
                f"internal error: 'DataArray' object expected,"
                f" but got type {result.__class__.__name__!r}"
            )

        return result


def _format_callable(name: str, fn: Union[np.ufunc, Callable]):
    if name == "where":
        return "where(C,X,Y)"
    if isinstance(fn, np.ufunc):
        num_args = fn.nargs - 1
    else:
        signature = inspect.signature(fn)
        num_args = len(signature.parameters.values())
    if num_args == 0:
        return f"{name}()"
    elif num_args == 1:
        return f"{name}(X)"
    elif num_args == 2:
        return f"{name}(X,Y)"
    elif num_args == 3:
        return f"{name}(X,Y,Z)"
    else:
        return f"{name}({','.join(map(lambda i: f'X{i + 1}', range(num_args)))})"



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import (
    JsonBooleanSchema,
    JsonIntegerSchema,
    JsonObjectSchema,
    JsonStringSchema,
)

from ..accessor import COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES, FsAccessor


class FileFsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return "file"

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                auto_mkdirs=JsonBooleanSchema(
                    description="Whether, when opening a file, the directory"
                    " containing it should be created (if it"
                    " doesn't already exist)."
                ),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            ),
            additional_properties=True,
        )


class MemoryFsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return "memory"


class S3FsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return "s3"

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        # We may use here AWS S3 defaults as described in
        #   https://boto3.amazonaws.com/v1/documentation/api/
        #   latest/guide/configuration.html
        return JsonObjectSchema(
            properties=dict(
                anon=JsonBooleanSchema(
                    title="Whether to anonymously connect to AWS S3."
                ),
                key=JsonStringSchema(
                    min_length=1,
                    title="AWS access key identifier.",
                    description="Can also be set in profile section"
                    " of ~/.aws/config, or by environment"
                    " variable AWS_ACCESS_KEY_ID.",
                ),
                secret=JsonStringSchema(
                    min_length=1,
                    title="AWS secret access key.",
                    description="Can also be set in profile section"
                    " of ~/.aws/config, or by environment"
                    " variable AWS_SECRET_ACCESS_KEY.",
                ),
                token=JsonStringSchema(
                    min_length=1,
                    title="Session token.",
                    description="Can also be set in profile section"
                    " of ~/.aws/config, or by environment"
                    " variable AWS_SESSION_TOKEN.",
                ),
                use_ssl=JsonBooleanSchema(
                    description="Whether to use SSL in connections to S3;"
                    " may be faster without, but insecure.",
                    default=True,
                ),
                requester_pays=JsonBooleanSchema(
                    description='If "RequesterPays" buckets are supported.',
                    default=False,
                ),
                s3_additional_kwargs=JsonObjectSchema(
                    description="parameters that are used when calling"
                    " S3 API methods. Typically used for"
                    ' things like "ServerSideEncryption".',
                    additional_properties=True,
                ),
                client_kwargs=JsonObjectSchema(
                    description="Parameters for the botocore client.",
                    properties=dict(
                        endpoint_url=JsonStringSchema(
                            min_length=1,
                            format="uri",
                            title="Alternative endpoint URL.",
                        ),
                        # bucket_name=JsonStringSchema(
                        #     min_length=1,
                        #     title='Name of the bucket'
                        # ),
                        profile_name=JsonStringSchema(
                            min_length=1,
                            title="Name of the AWS configuration profile",
                            description="Section name with within"
                            " ~/.aws/config file,"
                            " which provides AWS configurations"
                            " and credentials.",
                        ),
                        region_name=JsonStringSchema(
                            min_length=1, title="AWS storage region name"
                        ),
                    ),
                    additional_properties=True,
                ),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            ),
            additional_properties=True,
        )


class AzureFsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return "abfs"

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                anon=JsonBooleanSchema(
                    title="Whether to anonymously connect to Azure Blob Storage."
                ),
                account_name=JsonStringSchema(
                    min_length=1,
                    title="Azure storage account name.",
                    description="Must be used with the account key parameter."
                    " This is not required when using a"
                    " connection string.",
                ),
                account_key=JsonStringSchema(
                    min_length=1,
                    title="Azure storage account key.",
                    description="Must be used with the account"
                    " name parameter."
                    " This is not required when using a"
                    " connection string",
                ),
                connection_string=JsonStringSchema(
                    min_length=1,
                    title="Connection string for Azure blob storage.",
                    description="Use this parameter inplace of both"
                    " account name and key"
                    " because they are both contained"
                    " in the string.",
                ),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            ),
            additional_properties=True,
        )


class FtpFsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return "ftp"

    @classmethod
    def get_storage_options_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                host=JsonStringSchema(
                    title="FTP host",
                    description="The remote server name/ip to connect to",
                    min_length=1,
                ),
                port=JsonIntegerSchema(
                    minimum=0,
                    maximum=65535,
                    default=21,
                    title="FTP port",
                    description="Port to connect with",
                ),
                username=JsonStringSchema(
                    min_length=1,
                    title="User name",
                    description="If authenticating, the user's identifier",
                ),
                password=JsonStringSchema(
                    min_length=1,
                    title="User password",
                    description="User's password on the server, if using",
                ),
                **COMMON_STORAGE_OPTIONS_SCHEMA_PROPERTIES,
            ),
            additional_properties=True,
        )


class HttpsFsAccessor(FsAccessor):
    @classmethod
    def get_protocol(cls) -> str:
        return "https"



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from abc import ABCMeta, abstractmethod
from collections.abc import Collection, Mapping
from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import xarray as xr

from xcube.constants import CRS_WKT_EPSG_4326, EXTENSION_POINT_INPUT_PROCESSORS
from xcube.core.gridmapping import GridMapping
from xcube.core.resampling import rectify_dataset
from xcube.core.timecoord import (
    get_time_range_from_attrs,
    get_time_range_from_data,
    to_time_in_days_since_1970,
)
from xcube.util.plugin import ExtensionComponent, get_extension_registry


class ReprojectionInfo:
    """Characterize input datasets so we can reproject.

    Args:
        xy_names: Names of variables providing the spatial x- and
            y-coordinates, e.g. ('longitude', 'latitude')
        xy_tp_names: Optional names of tie-point variables providing the
            spatial y- and y-coordinates, e.g. ('TP_longitude',
            'TP_latitude')
        xy_crs: Optional spatial reference system, e.g. 'EPSG:4326' or
            WKT or proj4 mapping
        xy_gcp_step: Optional step size for collecting ground control
            points from spatial coordinate arrays denoted by *xy_names*.
        xy_tp_gcp_step: Optional step size for collecting ground control
            points from spatial coordinate arrays denoted by
            *xy_tp_names*.
    """

    def __init__(
        self,
        xy_names: tuple[str, str] = None,
        xy_tp_names: tuple[str, str] = None,
        xy_crs: Any = None,
        xy_gcp_step: Union[int, tuple[int, int]] = None,
        xy_tp_gcp_step: Union[int, tuple[int, int]] = None,
    ):
        self._xy_names = self._assert_name_pair("xy_names", xy_names)
        self._xy_tp_names = self._assert_name_pair("xy_tp_names", xy_tp_names)
        self._xy_crs = xy_crs
        self._xy_gcp_step = self._assert_step_pair("xy_gcp_step", xy_gcp_step)
        self._xy_tp_gcp_step = self._assert_step_pair("xy_tp_gcp_step", xy_tp_gcp_step)

    def derive(
        self,
        xy_names: tuple[str, str] = None,
        xy_tp_names: tuple[str, str] = None,
        xy_crs: Any = None,
        xy_gcp_step: Union[int, tuple[int, int]] = None,
        xy_tp_gcp_step: Union[int, tuple[int, int]] = None,
    ):
        return ReprojectionInfo(
            self.xy_names if xy_names is None else xy_names,
            xy_tp_names=self.xy_tp_names if xy_tp_names is None else xy_tp_names,
            xy_crs=self.xy_crs if xy_crs is None else xy_crs,
            xy_gcp_step=self.xy_gcp_step if xy_gcp_step is None else xy_gcp_step,
            xy_tp_gcp_step=(
                self.xy_tp_gcp_step if xy_tp_gcp_step is None else xy_tp_gcp_step
            ),
        )

    @property
    def xy_names(self) -> Optional[tuple[str, str]]:
        return self._xy_names

    @property
    def xy_tp_names(self) -> Optional[tuple[str, str]]:
        return self._xy_tp_names

    @property
    def xy_crs(self) -> Any:
        return self._xy_crs

    @property
    def xy_gcp_step(self) -> Optional[int]:
        return self._xy_gcp_step

    @property
    def xy_tp_gcp_step(self) -> Optional[int]:
        return self._xy_tp_gcp_step

    def _assert_name_pair(self, keyword: str, value):
        if value is not None:
            v1, v2 = value
            self._assert_name(keyword, v1)
            self._assert_name(keyword, v2)
            return v1, v2
        return value

    def _assert_step_pair(self, keyword: str, value):
        if value is not None:
            if isinstance(value, int):
                v1, v2 = value, value
            else:
                v1, v2 = value
            self._assert_step(keyword, v1)
            self._assert_step(keyword, v2)
            return v1, v2
        return value

    def _assert_name(self, keyword: str, value):
        if value is None:
            raise ValueError(f"invalid {keyword}, missing name")
        if not isinstance(value, str) or not value:
            raise ValueError(f"invalid {keyword}, name must be a non-empty string")

    def _assert_step(self, keyword: str, value):
        if value is None:
            raise ValueError(f"invalid {keyword}, missing name")
        if not isinstance(value, int) or value <= 0:
            raise ValueError(f"invalid {keyword}, step must be an integer number")


class InputProcessor(ExtensionComponent, metaclass=ABCMeta):
    """Read and process inputs for the gen tool.

    An InputProcessor can be configured by the following parameters:

    * ``input_reader``: The input format identifier. Required, no default.

    Args:
        name: A unique input processor identifier.
    """

    def __init__(self, name: str, **parameters):
        super().__init__(EXTENSION_POINT_INPUT_PROCESSORS, name)
        self._parameters = {**self.default_parameters, **parameters}
        if "input_reader" not in self._parameters:
            raise ValueError("missing input_reader in input processor parameters")

    @property
    def description(self) -> str:
        """Returns:
        A description for this input processor
        """
        return self.get_metadata_attr("description", "")

    @property
    def default_parameters(self) -> dict[str, Any]:
        return {}

    @property
    def parameters(self) -> Mapping[str, Any]:
        return self._parameters

    @property
    def input_reader(self) -> str:
        return self.parameters["input_reader"]

    @property
    def input_reader_params(self) -> dict:
        """Returns:
        The input reader parameters for this input processor.
        """
        return self.parameters.get("input_reader_params", {})

    @abstractmethod
    def get_time_range(self, dataset: xr.Dataset) -> Optional[tuple[float, float]]:
        """Return a tuple of two floats representing start/stop time (which may be same) in days since 1970.

        Args:
            dataset: The dataset.

        Returns:
            The time-range tuple of the dataset or None.
        """
        raise NotImplementedError()

    def get_extra_vars(self, dataset: xr.Dataset) -> Optional[Collection[str]]:
        """Get a set of names of variables that are required as input for the pre-processing and processing
        steps and should therefore not be dropped.
        However, the processing or post-processing steps may later remove them.

        Returns ``None`` by default.

        Args:
            dataset: The dataset.

        Returns:
            Collection of names of variables to be prevented from being
            dropping.
        """
        return None

    def pre_process(self, dataset: xr.Dataset) -> xr.Dataset:
        """Do any pre-processing before reprojection.
        All variables in the output dataset must be 2D arrays with dimensions "lat" and "lon", in this order.
        For example, perform dataset validation, masking, and/or filtering using provided configuration parameters.

        The default implementation returns the unchanged *dataset*.

        Args:
            dataset: The dataset.

        Returns:
            The pre-processed dataset or the original one, if no pre-
            processing is required.
        """
        return dataset

    # Note, arg "geo_coding" should be called "source_gm",
    # arg "output_geom" should be called "target_gm",
    # but this package will be deprecated anyway.
    # See #540

    @abstractmethod
    def process(
        self,
        dataset: xr.Dataset,
        geo_coding: GridMapping,
        output_geom: GridMapping,
        output_resampling: str,
        include_non_spatial_vars=False,
    ) -> xr.Dataset:
        """Perform spatial transformation into the cube's WGS84 SRS such that all variables in the output dataset
        * must be 2D arrays with dimensions "lat" and "lon", in this order, and
        * must have shape (*dst_size[-1]*, *dst_size[-2]*), and
        * must have *dst_region* as their bounding box in geographic coordinates.

        Args:
            dataset: The input dataset.
            geo_coding: The input's geo-coding.
            output_geom: The output's spatial image geometry.
            output_resampling: The output's spatial resampling method.
            include_non_spatial_vars: Whether to include non-spatial
                variables in the output.

        Returns:
            The transformed output dataset or the original one, if no
            transformation is required.
        """
        raise NotImplementedError()

    def post_process(self, dataset: xr.Dataset) -> xr.Dataset:
        """Do any post-processing transformation. The input is a 3D array with dimensions ("time", "lat", "lon").
        Post-processing may, for example, generate new "wavelength" dimension for variables whose name follow
        a certain pattern.

        The default implementation returns the unchanged *dataset*.

        Args:
            dataset: The dataset.

        Returns:
            The post-processed dataset or the original one, if no post-
            processing is required.
        """
        return dataset


class XYInputProcessor(InputProcessor, metaclass=ABCMeta):
    """Read and process inputs for the gen tool.

    An XYInputProcessor can be configured by the following parameters:

    * ``input_reader``: The input format identifier.
        Required, no default.
    * ``xy_names``: A tuple of names of the variable providing x,y geo-locations.
        Optional, looked up automatically if not given e.g. ``("lon", "lat")``.
    * ``xy_tp_names``: A tuple of names of the variable providing x,y tie-point geo-locations.
        Optional, no default.
    * ``xy_crs``: A WKT string that identifies the x,y coordinate reference system (CRS).
        Optional, no default.
    * ``xy_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    * ``xy_tp_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y tie-point coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    """

    @property
    def default_parameters(self) -> dict[str, Any]:
        default_parameters = super().default_parameters
        default_parameters.update(xy_names=("lon", "lat"))
        return default_parameters

    def get_reprojection_info(self, dataset: xr.Dataset) -> ReprojectionInfo:
        """Information about special fields in input datasets used for reprojection.

        Args:
            dataset: The dataset.

        Returns:
            The reprojection information of the dataset or None.
        """
        parameters = self.parameters
        return ReprojectionInfo(
            xy_names=parameters.get("xy_names", ("lon", "lat")),
            xy_tp_names=parameters.get("xy_tp_names"),
            xy_crs=parameters.get("xy_crs"),
            xy_gcp_step=parameters.get("xy_gcp_step"),
            xy_tp_gcp_step=parameters.get("xy_tp_gcp_step"),
        )

    def get_extra_vars(self, dataset: xr.Dataset) -> Optional[Collection[str]]:
        """Return the names of variables containing spatial coordinates.
        They should not be removed, as they are required for the reprojection.
        """
        reprojection_info = self.get_reprojection_info(dataset)
        if reprojection_info is None:
            return dataset
        extra_vars = set()
        if reprojection_info.xy_names:
            extra_vars.update(set(reprojection_info.xy_names))
        if reprojection_info.xy_tp_names:
            extra_vars.update(set(reprojection_info.xy_tp_names))
        return extra_vars

    def process(
        self,
        dataset: xr.Dataset,
        geo_coding: GridMapping,
        output_geom: GridMapping,
        output_resampling: str,
        include_non_spatial_vars=False,
    ) -> xr.Dataset:
        """Perform reprojection using tie-points / ground control points."""
        reprojection_info = self.get_reprojection_info(dataset)

        warn_prefix = "unsupported argument in np-GCP rectification mode"
        if reprojection_info.xy_crs is not None:
            warnings.warn(
                f"{warn_prefix}: ignoring "
                f"reprojection_info.xy_crs = {reprojection_info.xy_crs!r}"
            )
        if reprojection_info.xy_tp_names is not None:
            warnings.warn(
                f"{warn_prefix}: ignoring "
                f"reprojection_info.xy_tp_names = {reprojection_info.xy_tp_names!r}"
            )
        if reprojection_info.xy_gcp_step is not None:
            warnings.warn(
                f"{warn_prefix}: ignoring "
                f"reprojection_info.xy_gcp_step = {reprojection_info.xy_gcp_step!r}"
            )
        if reprojection_info.xy_tp_gcp_step is not None:
            warnings.warn(
                f"{warn_prefix}: ignoring "
                f"reprojection_info.xy_tp_gcp_step = {reprojection_info.xy_tp_gcp_step!r}"
            )
        if output_resampling != "Nearest":
            warnings.warn(
                f"{warn_prefix}: ignoring dst_resampling = {output_resampling!r}"
            )
        if include_non_spatial_vars:
            warnings.warn(
                f"{warn_prefix}: ignoring "
                f"include_non_spatial_vars = {include_non_spatial_vars!r}"
            )

        geo_coding = geo_coding.derive(
            xy_var_names=(reprojection_info.xy_names[0], reprojection_info.xy_names[1])
        )

        dataset = rectify_dataset(
            dataset, compute_subset=False, source_gm=geo_coding, target_gm=output_geom
        )
        if output_geom.is_tiled:
            # The following condition may become true,
            # if we have used rectified_dataset(input, ..., is_y_reverse=True)
            # In this case y-chunksizes will also be reversed. So that the first chunk is smaller than any other.
            # Zarr will reject such datasets, when written.
            if dataset.chunks.get("lat")[0] < dataset.chunks.get("lat")[-1]:
                dataset = dataset.chunk(
                    {"lat": output_geom.tile_height, "lon": output_geom.tile_width}
                )
        if (
            dataset is not None
            and geo_coding.crs.is_geographic
            and geo_coding.xy_var_names != ("lon", "lat")
        ):
            dataset = dataset.rename(
                {geo_coding.xy_var_names[0]: "lon", geo_coding.xy_var_names[1]: "lat"}
            )

        return dataset


class DefaultInputProcessor(XYInputProcessor):
    """Default input processor that expects input datasets to have the xcube standard format:

    * Have dimensions ``lat``, ``lon``, optionally ``time`` of length 1;
    * have coordinate variables ``lat[lat]``, ``lon[lat]``, ``time[time]`` (opt.), ``time_bnds[time, 2]`` (opt.);
    * have coordinate variables ``lat[lat]``, ``lon[lat]`` as decimal degrees on WGS84 ellipsoid,
      both linearly increasing with same constant delta;
    * have coordinate variable ``time[time]`` representing a date+time values with defined CF "units" attribute;
    * have any data variables of form ``<var>[time, lat, lon]``;
    * have global attribute pairs (``time_coverage_start``, ``time_coverage_end``), or (``start_time``, ``stop_time``)
      if ``time`` coordinate is missing.

    The default input processor can be configured by the following parameters:

    * ``input_reader``: The input format identifier.
        Required, defaults to ``"netcdf4"``.
    * ``xy_names``: A tuple of names of the variable providing x,y geo-locations.
        Optional, defaults to ``("lon", "lat")``.
    * ``xy_tp_names``: A tuple of names of the variable providing x,y tie-point geo-locations.
        Optional, no default.
    * ``xy_crs``: A WKT string that identifies the x,y coordinate reference system (CRS).
        Optional, defaults to WKT for EPSG:4326 (see ``xcube.constants.CRS_WKT_EPSG_4326`` constant).
    * ``xy_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    * ``xy_tp_gcp_step``: An integer or tuple of integers that is used to sub-sample x,y tie-point coordinate variables
        for extracting ground control points (GCP).
        Optional, no default.
    """

    def __init__(self, **parameters):
        super().__init__("default", **parameters)

    @property
    def default_parameters(self) -> dict[str, Any]:
        default_parameters = super().default_parameters
        default_parameters.update(
            input_reader="netcdf4", xy_names=("lon", "lat"), xy_crs=CRS_WKT_EPSG_4326
        )
        return default_parameters

    def pre_process(self, dataset: xr.Dataset) -> xr.Dataset:
        self._validate(dataset)

        if "time" in dataset.sizes:
            # Remove time dimension of length 1.
            dataset = dataset.squeeze("time")

        return _normalize_lon_360(dataset)

    def get_time_range(self, dataset: xr.Dataset) -> tuple[float, float]:
        time_coverage_start, time_coverage_end = get_time_range_from_data(dataset)
        if time_coverage_start is not None:
            time_coverage_start = str(time_coverage_start)
        if time_coverage_end is not None:
            time_coverage_end = str(time_coverage_end)
        if time_coverage_start is None or time_coverage_end is None:
            time_coverage_start, time_coverage_end = get_time_range_from_attrs(dataset)
        if time_coverage_start is None:
            raise ValueError(
                "invalid input: missing time coverage information in dataset"
            )
        if time_coverage_end is None:
            time_coverage_end = time_coverage_start
        return to_time_in_days_since_1970(
            time_coverage_start
        ), to_time_in_days_since_1970(time_coverage_end)

    def _validate(self, dataset):
        self._check_coordinate_var(dataset, "lon", min_length=2)
        self._check_coordinate_var(dataset, "lat", min_length=2)
        if "time" in dataset.sizes:
            self._check_coordinate_var(dataset, "time", max_length=1)
            required_dims = ("time", "lat", "lon")
        else:
            required_dims = ("lat", "lon")
        count = 0
        for var_name in dataset.data_vars:
            var = dataset.data_vars[var_name]
            if var.dims == required_dims:
                count += 1
        if count == 0:
            raise ValueError(
                f"dataset has no variables with required dimensions {required_dims!r}"
            )

    # noinspection PyMethodMayBeStatic
    def _check_coordinate_var(
        self,
        dataset: xr.Dataset,
        coord_var_name: str,
        min_length: int = None,
        max_length: int = None,
    ):
        if coord_var_name not in dataset.coords:
            raise ValueError(f'missing coordinate variable "{coord_var_name}"')
        coord_var = dataset.coords[coord_var_name]
        if len(coord_var.shape) != 1:
            raise ValueError('coordinate variable "lon" must be 1D')
        coord_var_bnds_name = coord_var.attrs.get("bounds", coord_var_name + "_bnds")
        if coord_var_bnds_name in dataset:
            coord_bnds_var = dataset[coord_var_bnds_name]
            expected_shape = (len(coord_var), 2)
            if coord_bnds_var.shape != expected_shape:
                raise ValueError(
                    f'coordinate bounds variable "{coord_bnds_var}" must have shape {expected_shape!r}'
                )
        else:
            if min_length is not None and len(coord_var) < min_length:
                raise ValueError(
                    f'coordinate variable "{coord_var_name}" must have at least {min_length} value(s)'
                )
            if max_length is not None and len(coord_var) > max_length:
                raise ValueError(
                    f'coordinate variable "{coord_var_name}" must have no more than {max_length} value(s)'
                )


def _normalize_lon_360(dataset: xr.Dataset) -> xr.Dataset:
    """Fix the longitude of the given dataset ``dataset`` so that it ranges from -180 to +180 degrees.

    Args:
        dataset: The dataset whose longitudes may be given in the range
            0 to 360.

    Returns:
        The fixed dataset or the original dataset.
    """

    if "lon" not in dataset.coords:
        return dataset

    lon_var = dataset.coords["lon"]

    if len(lon_var.shape) != 1:
        return dataset

    lon_size = lon_var.shape[0]
    if lon_size < 2:
        return dataset

    lon_size_05 = lon_size // 2
    lon_values = lon_var.values
    if not np.any(lon_values[lon_size_05:] > 180.0):
        return dataset

    # roll_coords will be set to False by default in the future
    dataset = dataset.roll(lon=lon_size_05, roll_coords=True)
    dataset = dataset.assign_coords(lon=(((dataset.lon + 180) % 360) - 180))

    return dataset


def find_input_processor_class(name: str):
    extension = get_extension_registry().get_extension(
        EXTENSION_POINT_INPUT_PROCESSORS, name
    )
    if not extension:
        return None
    return extension.component



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os
import os.path
from typing import Callable

import click

from xcube.cli.common import (
    cli_option_quiet,
    cli_option_verbosity,
    configure_cli_output,
)
from xcube.constants import FORMAT_NAME_ZARR, LOG, LOG_LEVEL_DETAIL

Monitor = Callable[[str, int], None]


# noinspection PyShadowingBuiltins
@click.command(name="prune")
@click.argument("dataset_path", metavar="DATASET")
@cli_option_quiet
@cli_option_verbosity
@click.option(
    "--dry-run",
    is_flag=True,
    help="Just read and process input, but don't produce any output.",
)
def prune(dataset_path: str, quiet: bool, verbosity: int, dry_run: bool):
    """
    Delete empty chunks.
    Deletes all data files associated with empty
    (NaN-only) chunks in given DATASET,
    which must have Zarr format.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    def monitor(msg: str, monitor_verbosity: int = 1):
        if monitor_verbosity == 0:
            LOG.error(msg)
        elif monitor_verbosity == 1:
            LOG.info(msg)
        elif monitor_verbosity == 2:
            LOG.log(LOG_LEVEL_DETAIL, msg)
        elif monitor_verbosity == 3:
            LOG.debug(msg)

    _prune(input_path=dataset_path, dry_run=dry_run, monitor=monitor)
    return 0


def _prune(input_path: str, dry_run: bool, monitor: Monitor):
    from xcube.core.chunk import get_empty_dataset_chunks
    from xcube.core.dsio import guess_dataset_format, open_dataset

    input_format = guess_dataset_format(input_path)
    if input_format != FORMAT_NAME_ZARR:
        raise click.ClickException("input must be a dataset in Zarr format")

    num_deleted_total = 0

    monitor(f"Opening dataset from {input_path!r}...", 1)
    with open_dataset(input_path) as dataset:
        monitor("Identifying empty chunks...", 1)
        for var_name, chunk_indices in get_empty_dataset_chunks(dataset):
            num_empty_chunks = 0
            num_deleted = 0
            for chunk_index in chunk_indices:
                num_empty_chunks += 1
                if num_empty_chunks == 1:
                    monitor(
                        f"Found empty chunks in variable {var_name!r}, "
                        f"deleting block files...",
                        2,
                    )

                ok = _delete_block_file(
                    input_path, var_name, chunk_index, dry_run, monitor
                )
                if ok:
                    num_deleted += 1
            if num_deleted > 0:
                monitor(
                    f"Deleted {num_deleted} block file(s) for variable {var_name!r}.",
                    2,
                )
            elif num_empty_chunks > 0:
                monitor(
                    f"No block files for variable {var_name!r} could be deleted.", 2
                )
            num_deleted_total += num_deleted

    monitor(f"Done, {num_deleted_total} block file(s) deleted total.", 1)


def _delete_block_file(
    input_path: str, var_name: str, chunk_index, dry_run: bool, monitor: Monitor
) -> bool:
    block_path = None
    block_path_1 = os.path.join(input_path, var_name, ".".join(map(str, chunk_index)))
    block_path_2 = block_path_1

    if os.path.isfile(block_path_1):
        block_path = block_path_1
    else:
        block_path_2 = os.path.join(input_path, var_name, *map(str, chunk_index))
        if os.path.isfile(block_path_2):
            block_path = block_path_2

    if block_path:
        monitor(f"Deleting chunk file {block_path}", 3)
        if dry_run:
            return True
        try:
            os.remove(block_path)
            return True
        except OSError as e:
            monitor(f"Failed to delete block file {block_path}: {e}", 0)
    else:
        monitor(f"Could find neither block file {block_path_1} nor {block_path_2}", 0)
    return False



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.schema import rechunk_cube

from ..config import CubeConfig
from .transformer import CubeTransformer, TransformedCube


class CubeRechunker(CubeTransformer):
    """Force cube to have chunks compatible with Zarr."""

    def transform_cube(
        self, cube: xr.Dataset, gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        cube, gm = rechunk_cube(
            cube, gm, chunks=cube_config.chunks, tile_size=cube_config.tile_size
        )
        return cube, gm, cube_config



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import importlib
import inspect
import os.path
import sys
import warnings
from typing import Any, Callable, Optional, Union

from xcube.util.assertions import assert_given, assert_instance, assert_true
from xcube.util.jsonschema import (
    JsonBooleanSchema,
    JsonObject,
    JsonObjectSchema,
    JsonStringSchema,
)
from xcube.util.temp import new_temp_dir

from ...constants import LOG
from .constants import DEFAULT_CALLABLE_NAME, DEFAULT_MODULE_NAME, TEMP_FILE_PREFIX
from .fileset import FileSet


class CodeConfig(JsonObject):
    """Code configuration object.

    Instances should always be created using one of the factory methods:

    * :meth:`from_callable`
    * :meth:`from_code`
    * :meth:`from_file_set`

    Args:
        _callable: Optional function or class. Cannot be given if
            *inline_code* or *file_set* are given.
        callable_ref: Optional reference to the callable in the
            *file_set*. Must have form "<module-name>:<callable-name>".
        inline_code: Optional inline Python code string. Cannot be given
            if *_callable* or *file_set* are given.
        file_set: A file set that contains Python modules or packages.
            Must be of type :class:`FileSet`. Cannot be given if
            *_callable* or *inline_code* are given.
        install_required: Whether *file_set* contains Python modules or
            packages that must be installed. Can only be True if
            *file_set* is given.
        callable_params: Optional parameters that are supposed to be
            passed as keyword-arguments to the callable. Stored as part
            of the configuration, but this class does not apply them.
            Must be a dictionary, if given.
    """

    def __init__(
        self,
        _callable: Callable = None,
        callable_ref: str = None,
        callable_params: dict[str, Any] = None,
        inline_code: str = None,
        file_set: FileSet = None,
        install_required: bool = None,
    ):
        if callable_ref is not None:
            # Will validate callable_ref
            _normalize_callable_ref(callable_ref)
        if _callable is not None:
            assert_true(callable(_callable), "_callable must be a function or class")
        if callable_params is not None:
            assert_instance(callable_params, dict, "callable_params")
        if inline_code is not None:
            assert_instance(inline_code, str, "inline_code")
        if file_set is not None:
            assert_instance(file_set, FileSet, "file_set")
        assert_true(
            sum((1 if v is not None else 0) for v in (_callable, inline_code, file_set))
            == 1,
            "_callable, inline_code, file_set "
            "are mutually exclusive and one must be given",
        )
        assert_true(
            not install_required or file_set is not None,
            "install_required can only be used if file_set is given",
        )
        self._callable = _callable
        self.callable_ref = callable_ref
        self.callable_params = callable_params
        self.inline_code = inline_code
        self.file_set = file_set
        self.install_required = install_required

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        """Get the JSON schema for CodeConfig objects."""
        return JsonObjectSchema(
            properties=dict(
                callable_ref=JsonStringSchema(min_length=1),
                callable_params=JsonObjectSchema(additional_properties=True),
                inline_code=JsonStringSchema(min_length=1),
                file_set=FileSet.get_schema(),
                install_required=JsonBooleanSchema(),
            ),
            additional_properties=False,
            factory=cls,
        )

    @classmethod
    def from_code(
        cls,
        *code: Union[str, Callable],
        callable_name: str = None,
        module_name: str = None,
        callable_params: dict[str, Any] = None,
    ) -> "CodeConfig":
        """Create a code configuration from the given *code* which may be
        given as one or more plain text strings or callables.

        This will create a configuration that uses an inline
        ``code_string`` which contains the source code.

        Args:
            *code: The code.
            callable_name: The callable name. If not given, will be
                inferred from first callable. Otherwise, it defaults to
                "process_dataset".
            module_name: The module name. If not given, defaults to
                "user_code".
            callable_params: Optional parameters that are supposed to be
                passed as keyword-arguments to the callable. Stored as
                part of the configuration, but this class does not apply
                them. Must be a dictionary, if given.

        Returns:
            A new code configuration.
        """
        assert_given(code, "code")
        inline_code, callable_ref = _normalize_inline_code(
            code, module_name=module_name, callable_name=callable_name
        )
        return CodeConfig(
            callable_ref=callable_ref,
            inline_code=inline_code,
            callable_params=callable_params,
        )

    @classmethod
    def from_callable(
        cls, _callable: Callable, callable_params: Optional[dict[str, Any]] = None
    ) -> "CodeConfig":
        """Create a code configuration from the callable *_callable*.

        Note, the resulting code configuration is only
        valid in a local context. It cannot be JSON-serialized.

        To pass such configurations to a service, convert it first
        using the :meth:`to_service` first.

        Args:
            _callable: A function or class
            callable_params: Optional parameters that are supposed to be
                passed as keyword-arguments to the callable. Stored as
                part of the configuration, but this class does not apply
                them. Must be a dictionary, if given.

        Returns:
            A new code configuration.
        """
        assert_given(_callable, "_callable")
        return CodeConfig(_callable=_callable, callable_params=callable_params)

    @classmethod
    def from_file_set(
        cls,
        file_set: Union[FileSet, str, Any],
        callable_ref: str,
        callable_params: Optional[dict[str, Any]] = None,
        install_required: Optional[bool] = None,
    ) -> "CodeConfig":
        """Create a code configuration from a file set.

        Args:
            file_set: The file set. Can be a path or a :class:FileSet
                instance.
            callable_ref: Reference to the callable in the *file_set*,
                must have form "<module-name>:<callable-name>"
            install_required: Whether the *file_set* is a package that
                must be installed.
            callable_params: Optional parameters that are supposed to be
                passed as keyword-arguments to the callable. Stored as
                part of the configuration, but this class does not apply
                them. Must be a dictionary, if given.

        Returns:
            A new code configuration.
        """
        assert_given(file_set, "file_set")
        assert_given(callable_ref, "callable_ref")
        return CodeConfig(
            callable_ref=callable_ref,
            file_set=_normalize_file_set(file_set),
            install_required=install_required,
            callable_params=callable_params,
        )

    @classmethod
    def from_github_archive(
        cls,
        gh_org: str,
        gh_repo: str,
        gh_tag: str,
        gh_release: str,
        callable_ref: str,
        callable_params: Optional[dict[str, Any]] = None,
        gh_username: Optional[str] = None,
        gh_token: Optional[str] = None,
    ):
        """Create a code configuration from a GitHub archive.

        Args:
            gh_org: GitHub organisation name or username
            gh_repo: GitHub repository name
            gh_tag: GitHub release tag
            gh_release: The name of a GitHub release. It is used to form
                the sub-path into the archive. The sub-path has the form
                "<gh_repo>-<gh_release>".
            callable_ref: Reference to the callable in the *file_set*,
                must have form "<module-name>:<callable-name>"
            callable_params: Optional parameters that are supposed to be
                passed as keyword-arguments to the callable. Stored as
                part of the configuration, but this class does not apply
                them. Must be a dictionary, if given.
            gh_username: Optional GitHub username.
            gh_token: Optional GitHub username.

        Returns:

        """
        assert_given(gh_org, "gh_org")
        assert_given(gh_org, "gh_repo")
        assert_given(gh_org, "gh_tag")
        assert_given(gh_org, "gh_release")
        gh_url = f"https://github.com/{gh_org}/{gh_repo}/archive/{gh_tag}.zip"
        gh_sub_path = f"{gh_repo}-{gh_release}"
        gh_params = None
        if gh_token is not None:
            gh_params = dict(token=gh_token, username=gh_username or gh_org)
        elif gh_username is not None:
            assert_given(gh_token, "gh_token")  # fails always
        return CodeConfig(
            file_set=FileSet(gh_url, sub_path=gh_sub_path, storage_params=gh_params),
            callable_ref=callable_ref,
            callable_params=callable_params,
        )

    def for_service(self) -> "CodeConfig":
        """Convert this code configuration so can be used by the
        generator service.

        The returned code configuration defines either
        inline code or a file set that represents a local
        ZIP archive.
        """
        return _for_service(self)

    def for_local(self) -> "CodeConfig":
        """Convert this code configuration so can be used by the
        local generator. This means, the returned configuration
        can be used to load executable code. i.e.
        :meth:`get_callable` will return a callable.

        The returned code configuration defines either
        a callable (member _callable) or a file set that represents
        a local directory. This directory may be used without
        package installation (i.e. added to Python's sys.path) or
        used after installation (i.e. by executing
        ``pip install .``).
        """
        return _for_local(self)

    def get_callable(self) -> Callable:
        """Get the callable specified by this configuration.

        In the common case, this will require importing the callable.

        Returns:
            A callable

        Raises:
            ImportError: if the callable can not be imported
        """
        if self._callable is None:
            self.set_callable(self._load_callable())
        return self._callable

    def set_callable(self, func_or_class: Callable):
        """Set the callable that is represented by this configuration.

        Args:
            func_or_class: A callable
        """
        assert_true(callable(func_or_class), f"func_or_class must be callable")
        self._callable = func_or_class

    def _load_callable(self) -> Callable:
        """Load the callable specified by this configuration.

        Returns:
            A callable

        Raises:
            ImportError: if the callable can not be imported
        """
        code_config = self.for_local()
        return _load_callable(
            code_config.file_set.path,
            code_config.callable_ref,
            code_config.install_required,
        )


def _for_service(code_config: CodeConfig) -> "CodeConfig":
    if code_config.inline_code is not None:
        # If we have inline code,
        # there is no need to do anything else.
        return code_config

    if code_config.file_set is not None and not code_config.file_set.is_local():
        # For service requests, remote file set will
        # simply turn into JSON. for_local() will download
        # them when request is executed.
        # So we are done here.
        return code_config

    # noinspection PyProtectedMember
    if (
        code_config._callable is not None
        and code_config.inline_code is None
        and code_config.file_set is None
    ):
        # If we have callable that was not loaded
        # from inline string or files set,
        # turn into a local Python module.
        # noinspection PyProtectedMember
        code_config = _callable_to_module(
            code_config._callable, code_config.callable_params
        )

    if code_config.file_set is not None and code_config.file_set.is_local_dir():
        # If we have a file set that is a local
        # directory, turn it into a local ZIP Archive.
        file_set = code_config.file_set.to_local_zip()
        code_config = code_config.from_file_set(
            file_set=file_set,
            callable_ref=code_config.callable_ref,
            callable_params=code_config.callable_params,
            install_required=code_config.install_required,
        )

    if code_config.file_set is not None and code_config.file_set.is_local_zip():
        # Local ZIP archives will become file-attachments
        # to HTTP service requests. In the request's JSON we will
        # still have the local path.
        # So we are done here.
        return code_config

    if code_config.file_set is not None and code_config.file_set.is_local_zip():
        # At this point, code_config.file_set
        # is always a local ZIP archive.
        return code_config

    raise RuntimeError("for_service() failed due to an invalid CodeConfig state")


def _for_local(code_config: CodeConfig) -> "CodeConfig":
    # noinspection PyProtectedMember
    if code_config._callable is not None:
        # If the callable is already defined,
        # there is no need to do anything else.
        return code_config

    if code_config.inline_code:
        # Turn inline code into a local Python module.
        code_config = _inline_code_to_module(
            code_config.inline_code,
            code_config.callable_ref,
            code_config.callable_params,
        )
    file_set = code_config.file_set
    if file_set is None:
        raise RuntimeError(
            "CodeConfig.for_local() failed due to an invalid internal state"
        )

    if file_set.is_local_dir() or (file_set.is_local_zip() and not file_set.sub_path):
        # Either a directory or Zip file that can be added
        # to sys.path as is
        return code_config

    # If the file set is not a local directory,
    # turn it into one.
    return code_config.from_file_set(
        file_set=file_set.to_local_dir(),
        callable_ref=code_config.callable_ref,
        install_required=code_config.install_required,
        callable_params=code_config.callable_params,
    )


def _load_callable(
    dir_path: str, callable_ref: str, install_required: bool
) -> Callable:
    """Load the callable from given *dir_path* using the
    callable reference *callable_red*.

    Args:
        dir_path: A local directory path (local ZIPs should work too).
        callable_ref: A callable reference of form "module:callable"
        install_required: Whether source code in *dir_path* must be
            installed using setup.

    Returns:
        The loaded callable
    """
    module_name, callable_name = _normalize_callable_ref(callable_ref)
    if install_required:
        warnings.warn(
            f"This user-code configuration requires "
            f"package installation, "
            f"but this is not supported yet"
        )
    if dir_path not in sys.path:
        # Ok, we need to change global state here.
        # This cannot be done temporarily (e.g., by using a context manager)
        # as additional code might be dynamically loaded later
        # from local imports.
        LOG.warning(f"Python sys.path prepended by {dir_path}")
        sys.path = [dir_path] + sys.path
    # Now we should be able to import the module...
    # (import_module() will raise an ImportError otherwise).
    module = importlib.import_module(module_name)
    # ...and find the callable.
    _callable = getattr(module, callable_name, None)
    if _callable is None:
        raise ImportError(
            f"callable {callable_name!r} not found in module {module_name!r}"
        )
    if not callable(_callable):
        raise ImportError(f"{callable_name} of module {callable_name} is not callable")
    LOG.info(f"Imported {callable_ref!r} from {dir_path}")
    return _callable


_user_module_counter = 0


def _next_user_module_name() -> str:
    global _user_module_counter
    _user_module_counter += 1
    return f"{DEFAULT_MODULE_NAME}_{_user_module_counter}"


def _callable_to_module(
    _callable: Callable, callable_params: dict[str, Any] = None
) -> CodeConfig:
    """Create a code configuration from the callable *_callable*.

    This will create a configuration that uses a ``file_set``
    which contains the source code for the *func_or_class*.

    Args:
        _callable: A function or class
        callable_params: The parameters passed as keyword-arguments to
            the callable.

    Returns:
        A new code configuration.
    """
    callable_name = _callable.__name__
    if not callable_name:
        raise ValueError(f"cannot detect name for func_or_class")

    module = inspect.getmodule(_callable)
    module_name = module.__name__ if module is not None else None
    if not module_name:
        raise ValueError(f"cannot detect module for callable {callable_name!r}")

    source_file = inspect.getabsfile(_callable)
    if source_file is None:
        raise ValueError(
            f"cannot detect source file for func_or_class {callable_name!r}"
        )

    module_path, ext = os.path.splitext(os.path.normpath(source_file))
    if not module_path.replace(os.path.sep, "/").endswith(
        module_name.replace(".", "/")
    ):
        raise ValueError(
            f"cannot detect module path for func_or_class {callable_name!r}"
        )

    module_path = os.path.normpath(module_path[0 : -len(module_name)])

    code_config = CodeConfig.from_file_set(
        file_set=FileSet(module_path, includes=["*.py"]),
        callable_ref=f"{module_name}:{callable_name}",
        callable_params=callable_params,
    )
    code_config.set_callable(_callable)
    return code_config


def _inline_code_to_module(
    inline_code: str, callable_ref: str, callable_params: dict[str, Any] = None
):
    module_name, callable_name = _normalize_callable_ref(callable_ref)
    dir_path = new_temp_dir(prefix=TEMP_FILE_PREFIX)
    with open(os.path.join(dir_path, f"{module_name}.py"), "w") as stream:
        stream.write(inline_code)
    return CodeConfig.from_file_set(
        file_set=FileSet(dir_path),
        callable_ref=callable_ref,
        callable_params=callable_params,
    )


def _normalize_file_set(file_set: Union[FileSet, str, Any]) -> FileSet:
    if isinstance(file_set, FileSet):
        return file_set
    return FileSet(str(file_set))


def _normalize_inline_code(
    code: tuple[Union[str, Callable], ...],
    callable_name: str = None,
    module_name: str = None,
) -> tuple[str, str]:
    if not callable_name:
        first_callable = next(filter(callable, code), None)
        if first_callable is not None:
            callable_name = first_callable.__name__
    callable_name = callable_name or DEFAULT_CALLABLE_NAME
    module_name = module_name or _next_user_module_name()
    if all(map(callable, code)):
        warnings.warn(
            "inline code may not be executable due to missing import statements"
        )
    inline_code = "\n\n".join(
        [c if isinstance(c, str) else inspect.getsource(c) for c in code]
    )
    return inline_code, f"{module_name}:{callable_name}"


def _normalize_callable_ref(callable_ref: str) -> tuple[Optional[str], str]:
    """Normalize a callable reference *callable_ref* of the form
    "[module_name:]callable_name" into a pair.
    """
    assert_instance(callable_ref, str, "callable_ref")

    splits = callable_ref.split(":")
    if len(splits) == 2:
        module_name, callable_name = (s.strip() for s in splits)
    else:
        module_name, callable_name = "", ""

    # Verify components

    assert_true(
        module_name and callable_name,
        message=f"illegal callable_ref {callable_ref!r}: "
        f'expected format "<module>:<callable>"',
    )
    assert_true(
        all(s.isidentifier() for s in module_name.split(".")),
        message=f"illegal callable_ref {callable_ref!r}: "
        f"{module_name!r} is not a valid Python module name",
    )
    assert_true(
        callable_name.isidentifier(),
        message=f"illegal callable_ref {callable_ref!r}: "
        f"{callable_name!r} is not a valid Python identifier",
    )

    return module_name, callable_name



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os.path
import shutil
from collections.abc import Sequence
from typing import Type, Union

import zarr

from xcube.core.unchunk import unchunk_dataset


def optimize_dataset(
    input_path: str,
    output_path: str = None,
    in_place: bool = False,
    unchunk_coords: Union[bool, str, Sequence[str]] = False,
    exception_type: type[Exception] = ValueError,
):
    """Optimize a dataset for faster access.

    Reduces the number of metadata and coordinate data files in xcube dataset given by given by *dataset_path*.
    Consolidated cubes open much faster from remote locations, e.g. in object storage,
    because obviously much less HTTP requests are required to fetch initial cube meta
    information. That is, it merges all metadata files into a single top-level JSON file ".zmetadata".

    If *unchunk_coords* is given, it also removes any chunking of coordinate variables
    so they comprise a single binary data file instead of one file per data chunk.
    The primary usage of this function is to optimize data cubes for cloud object storage.
    The function currently works only for data cubes using Zarr format.
    *unchunk_coords* can be a name, or list of names of the coordinate variable(s) to be consolidated.
    If boolean ``True`` is used, coordinate all variables will be consolidated.

    Args:
        input_path: Path to input dataset with ZARR format.
        output_path: Path to output dataset with ZARR format. May
            contain "{input}" template string, which is replaced by the
            input path's file name without file name extension.
        in_place: Whether to modify the dataset in place. If False, a
            copy is made and *output_path* must be given.
        unchunk_coords: The name of a coordinate variable or a list of
            coordinate variables whose chunks should be consolidated.
            Pass ``True`` to consolidate chunks of all coordinate
            variables.
        exception_type: Type of exception to be used on value errors.
    """

    if not os.path.isfile(os.path.join(input_path, ".zgroup")):
        raise exception_type("Input path must point to ZARR dataset directory.")

    input_path = os.path.abspath(os.path.normpath(input_path))

    if in_place:
        output_path = input_path
    else:
        if not output_path:
            raise exception_type(f"Output path must be given.")
        if "{input}" in output_path:
            base_name, _ = os.path.splitext(os.path.basename(input_path))
            output_path = output_path.format(input=base_name)
        output_path = os.path.abspath(os.path.normpath(output_path))
        if os.path.exists(output_path):
            raise exception_type(f"Output path already exists.")

    if not in_place:
        shutil.copytree(input_path, output_path)

    if unchunk_coords:
        if isinstance(unchunk_coords, str):
            var_names = (unchunk_coords,)
        elif isinstance(unchunk_coords, bool):
            var_names = None
        else:
            var_names = tuple(unchunk_coords)
        unchunk_dataset(output_path, var_names=var_names, coords_only=True)

    zarr.convenience.consolidate_metadata(output_path)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from typing import Type


def register_json_formatter(cls: type, to_dict_method_name: str = "to_dict"):
    """TODO

    Args:
        cls
        to_dict_method_name

    Returns:

    """
    if not hasattr(cls, to_dict_method_name) or not callable(
        getattr(cls, to_dict_method_name)
    ):
        raise ValueError(f"{cls} must define a {to_dict_method_name}() method")

    try:
        import IPython
        import IPython.display

        if IPython.get_ipython() is not None:

            def obj_to_dict(obj):
                return getattr(obj, to_dict_method_name)()

            ipy_formatter = IPython.get_ipython().display_formatter.formatters[
                "application/json"
            ]
            ipy_formatter.for_type(cls, obj_to_dict)

    except ImportError:
        pass


def enable_asyncio():
    """Enable asyncio package to be executable in Jupyter Notebooks."""
    try:
        import IPython

        if IPython.get_ipython() is not None:
            try:
                import nest_asyncio

                nest_asyncio.apply()
            except ImportError:
                warnings.warn(
                    "nest-asyncio required to use asyncio in Jupyter Notebooks"
                )
    except ImportError:
        pass



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

import xarray as xr


def compute_dataset(ds, period="1W", incl_stdev=False):
    if incl_stdev:
        resample_obj = ds.resample(time=period)
        ds_mean = resample_obj.mean(dim="time")
        ds_std = resample_obj.std(dim="time").rename(
            name_dict={name: f"{name}_stdev" for name in ds.data_vars}
        )
        ds_merged = xr.merge([ds_mean, ds_std])
        ds_merged.attrs.update(ds.attrs)
        return ds_merged
    else:
        return ds.resample(time=period).mean(dim="time")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import unittest

from xcube.plugin import init_plugin
from xcube.util.extension import ExtensionRegistry


class PluginTest(unittest.TestCase):
    def test_standard_extensions_exist(self):
        ext_reg = ExtensionRegistry()

        init_plugin(ext_reg)

        self.assertTrue(ext_reg.has_extension("xcube.core.gen.iproc", "default"))

        self.assertTrue(ext_reg.has_extension("xcube.core.dsio", "zarr"))
        self.assertTrue(ext_reg.has_extension("xcube.core.dsio", "netcdf4"))
        self.assertTrue(ext_reg.has_extension("xcube.core.dsio", "csv"))
        self.assertTrue(ext_reg.has_extension("xcube.core.dsio", "mem"))

        self.assertTrue(ext_reg.has_extension("xcube.core.store", "file"))
        self.assertTrue(ext_reg.has_extension("xcube.core.store", "memory"))
        self.assertTrue(ext_reg.has_extension("xcube.core.store", "s3"))

        self.assertTrue(ext_reg.has_extension("xcube.cli", "compute"))
        self.assertTrue(ext_reg.has_extension("xcube.cli", "extract"))
        self.assertTrue(ext_reg.has_extension("xcube.cli", "gen"))
        self.assertTrue(ext_reg.has_extension("xcube.cli", "level"))
        self.assertTrue(ext_reg.has_extension("xcube.cli", "optimize"))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import pyproj.crs
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.resampling import resample_in_space

from ..config import CubeConfig
from .transformer import CubeTransformer, TransformedCube


class CubeResamplerXY(CubeTransformer):
    def transform_cube(
        self, cube: xr.Dataset, source_gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        target_gm = _compute_target_grid_mapping(cube_config, source_gm)

        if source_gm.is_close(target_gm):
            resampled_cube = cube
        else:
            resampled_cube = resample_in_space(
                cube, source_gm=source_gm, target_gm=target_gm
            )

        cube_config = cube_config.drop_props(["crs", "spatial_res", "bbox"])

        return resampled_cube, target_gm, cube_config


def _compute_target_grid_mapping(
    cube_config: CubeConfig, source_gm: GridMapping
) -> GridMapping:
    # assert_true(source_gm.is_regular, 'source_gm must be regular')

    target_crs = cube_config.crs
    target_bbox = cube_config.bbox
    target_spatial_res = cube_config.spatial_res

    if target_crs is None and target_bbox is None and target_spatial_res is None:
        # Nothing to do
        if source_gm.is_regular:
            return source_gm
        return source_gm.to_regular(tile_size=cube_config.tile_size)

    if target_spatial_res is not None:
        xy_res = (target_spatial_res, target_spatial_res)
    else:
        xy_res = source_gm.xy_res
    if target_bbox is not None:
        x_res, y_res = xy_res
        x_min, y_min, x_max, y_max = target_bbox
        xy_min = x_min, y_min
        size = round((x_max - x_min) / x_res), round((y_max - y_min) / y_res)
    else:
        xy_min = source_gm.x_min, source_gm.y_min
        size = source_gm.size
    if target_crs is not None:
        crs = pyproj.crs.CRS.from_string(target_crs)
    else:
        crs = source_gm.crs
    target_gm = GridMapping.regular(
        size=size,
        xy_min=xy_min,
        xy_res=xy_res,
        crs=crs,
        tile_size=source_gm.tile_size,
        is_j_axis_up=source_gm.is_j_axis_up,
    )
    return target_gm.derive(
        xy_var_names=source_gm.xy_var_names, xy_dim_names=source_gm.xy_dim_names
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import random
from collections.abc import Iterable
from typing import Any, Union

import dask.array as da
import matplotlib.colors
import numpy as np
import xarray as xr

# TODO: this would be useful to have in xarray:
#       >>> ds = xr.open_dataset("my/path/to/cf/netcdf.nc", decode_flags=True)
#       >>> ds.flag_mask_sets['quality_flags']


_UINT16_MAX = 65535


class MaskSet:
    """A set of mask variables derived from a variable *flag_var* with the following
    CF attributes:

      - One or both of `flag_masks` and `flag_values`
      - `flag_meanings` (always required)

    See https://cfconventions.org/Data/cf-conventions/cf-conventions-1.9/cf-conventions.html#flags
    for details on the use of these attributes.

    Each mask is represented by an `xarray.DataArray`, has the name of the flag,
    is of type `numpy.unit8`, and has the dimensions of the given *flag_var*.

    Args:
        flag_var: an `xarray.DataArray` that defines flag values. The CF
            attributes `flag_meanings` and one or both of `flag_masks`
            and `flag_values` are expected to exist and be valid.
    """

    def __init__(self, flag_var: xr.DataArray):
        flag_masks = flag_var.attrs.get("flag_masks")
        flag_values = flag_var.attrs.get("flag_values")
        if flag_masks is None and flag_values is None:
            raise ValueError(
                "One or both of the attributes "
                "'flag_masks' or 'flag_values' "
                "must be present and non-null in flag_var"
            )
        if flag_masks is not None:
            flag_masks = _convert_flag_var_attribute_value(flag_masks, "flag_masks")
        if flag_values is not None:
            flag_values = _convert_flag_var_attribute_value(flag_values, "flag_values")
        if "flag_meanings" not in flag_var.attrs:
            raise ValueError("flag_var must have the attribute 'flag_meanings'")
        flag_meanings = flag_var.attrs.get("flag_meanings")
        if not isinstance(flag_meanings, str):
            raise TypeError("attribute 'flag_meanings' of flag_var must be a string")

        flag_names = flag_meanings.split(" ")
        if flag_masks is not None and len(flag_names) != len(flag_masks):
            raise ValueError(
                "attributes 'flag_meanings' and 'flag_masks' are not corresponding"
            )
        if flag_values is not None and len(flag_names) != len(flag_values):
            raise ValueError(
                "attributes 'flag_meanings' and 'flag_values' are not corresponding"
            )

        flag_colors = flag_var.attrs.get("flag_colors")
        if isinstance(flag_colors, str):
            flag_colors = flag_colors.split(" ")
        elif not isinstance(flag_colors, (list, tuple)):
            flag_colors = None

        self._masks = {}
        self._flag_var = flag_var
        self._flag_names = flag_names
        self._flag_masks = flag_masks
        self._flag_values = flag_values
        self._flag_colors = flag_colors

        if flag_masks is None:
            flag_masks = [None] * len(flag_names)
        if flag_values is None:
            flag_values = [None] * len(flag_names)
        self._flags = dict(zip(flag_names, list(zip(flag_masks, flag_values))))

    @classmethod
    def is_flag_var(cls, var: xr.DataArray) -> bool:
        return "flag_meanings" in var.attrs and (
            "flag_masks" in var.attrs or "flag_values" in var.attrs
        )

    @classmethod
    def get_mask_sets(cls, dataset: xr.Dataset) -> dict[str, "MaskSet"]:
        """For each "flag" variable in given *dataset*, turn it into a ``MaskSet``,
        store it in a dictionary.

        Args:
            dataset: The dataset

        Returns:
            A mapping of flag names to ``MaskSet``. Will be empty if
            there are no flag variables in *dataset*.
        """
        masks = {}
        for var_name in dataset.variables:
            var = dataset[var_name]
            if cls.is_flag_var(var):
                masks[var_name] = MaskSet(var)
        return masks

    def _repr_html_(self):
        lines = [
            "<html>",
            "<table>",
            "<tr><th>Flag name</th><th>Mask</th><th>Value</th></tr>",
        ]

        for name, data in self._flags.items():
            mask, value = data
            lines.append(f"<tr><td>{name}</td><td>{mask}</td><td>{value}</td></tr>")

        lines.extend(["</table>", "</html>"])

        return "\n".join(lines)

    def keys(self) -> Iterable[str]:
        return self._flag_names

    def __len__(self):
        return len(self._flag_names)

    def __str__(self):
        return "{}({})".format(
            self._flag_var.name,
            ", ".join([f"{n}=({v[0]}, {v[1]})" for n, v in self._flags.items()]),
            # We explicitly unpack the tuple to get the __str__ of the elements,
            # rather than the __repr__. See https://peps.python.org/pep-3140/
        )

    def __dir__(self) -> Iterable[str]:
        return self._flag_names

    def __getattr__(self, name: str) -> Any:
        if name not in self._flags:
            raise AttributeError(name)
        return self.get_mask(name)

    def __getitem__(self, item):
        try:
            name = self._flag_names[item]
            if name not in self._flags:
                raise IndexError(item)
        except TypeError:
            name = item
            if name not in self._flags:
                raise KeyError(item)
        return self.get_mask(name)

    def __contains__(self, item):
        return item in self._flags

    def get_mask(self, flag_name: str):
        if flag_name not in self._flags:
            raise ValueError('invalid flag name "%s"' % flag_name)

        if flag_name in self._masks:
            return self._masks[flag_name]

        flag_var = self._flag_var
        flag_mask, flag_value = self._flags[flag_name]

        if flag_var.chunks is not None:
            ones_array = da.ones(flag_var.shape, dtype=np.uint8, chunks=flag_var.chunks)
        else:
            ones_array = np.ones(flag_var.shape, dtype=np.uint8)

        mask_var = xr.DataArray(
            ones_array, dims=flag_var.dims, name=flag_name, coords=flag_var.coords
        )
        if flag_mask is not None:
            if flag_var.dtype != flag_mask.dtype:
                flag_var = flag_var.astype(flag_mask.dtype)
            if flag_value is not None:
                mask_var = mask_var.where((flag_var & flag_mask) == flag_value, 0)
            else:
                mask_var = mask_var.where((flag_var & flag_mask) != 0, 0)
        else:
            if flag_var.dtype != flag_value.dtype:
                flag_var = flag_var.astype(flag_value.dtype)
            mask_var = mask_var.where(flag_var == flag_value, 0)

        self._masks[flag_name] = mask_var
        return mask_var

    def get_cmap(
        self, default: str = "viridis"
    ) -> tuple[matplotlib.colors.Colormap, Union[matplotlib.colors.BoundaryNorm, None]]:
        """Get a suitable color mapping for use with matplotlib.

        Args:
            default: Default color map name in case a color mapping
                cannot be created, e.g., ``flag_values`` are not defined or
                if the flag values are not in the range [0, 2**16 - 1).

        Returns:
            A suitable instance of ``matplotlib.colors.Colormap`` and
            the corresponding ``matplotlib.colors.BoundaryNorm`` if applicable
        """
        if self._flag_values is not None:
            flag_var = self._flag_var
            cmap_name = str(flag_var.name) if flag_var.name else "from_maskset"

            flag_values = self._flag_values
            flag_value_min = flag_values.min()
            flag_value_max = flag_values.max()

            if flag_value_min >= 0 and flag_value_max < _UINT16_MAX:
                flag_colors = self._flag_colors
                levels = np.append(flag_values, flag_values[-1] + 1)
                # Setup color list that directly maps a flag value into a color.
                if flag_colors is not None:
                    colors = [(0, 0, 0, 0)] * len(flag_values)
                    if len(flag_colors) != len(flag_values):
                        # Handle special case where no-data value is included in
                        # fill_values (e.g., ESA CCI Land Cover) in the hope that
                        # flag_colors and flag_values will then match:
                        flag_values, index_tracker = _sanitize_flag_values(
                            flag_var, flag_values
                        )
                        if len(flag_colors) != len(flag_values):
                            return matplotlib.colormaps.get_cmap(default), None

                        # Use given flag colors
                        for i, c in zip(index_tracker, flag_colors):
                            colors[i] = c
                    else:
                        # Use given flag colors
                        for i, c in enumerate(flag_colors):
                            colors[i] = c

                else:
                    # Use random colors so they are all different
                    colors = [(0, 0, 0, 0)] * (len(flag_values))
                    for i in range(len(flag_values)):
                        colors[i] = (
                            random.random(),
                            random.random(),
                            random.random(),
                        )
                cmap = matplotlib.colors.ListedColormap(colors, name=cmap_name)
                cmap.set_over((0, 0, 0, 0))
                cmap.set_under((0, 0, 0, 0))
                norm = matplotlib.colors.BoundaryNorm(levels, len(colors))
                return cmap, norm
        return matplotlib.colormaps.get_cmap(default), None


_MASK_DTYPES = (
    (2**8, np.uint8),
    (2**16, np.uint16),
    (2**32, np.uint32),
    (2**64, np.uint64),
)


def _convert_flag_var_attribute_value(attr_value, attr_name) -> np.ndarray:
    if isinstance(attr_value, str):
        err_msg = f'Invalid bit expression in value for {attr_name}: "{attr_value}"'
        masks = []
        max_mask = 0
        for s in attr_value.split(","):
            s = s.strip()
            pair = s.split("-")
            if len(pair) == 1:
                try:
                    mask = (1 << int(s[0:-1])) if s.endswith("b") else int(s)
                except ValueError as e:
                    raise ValueError(err_msg) from e
            elif len(pair) == 2:
                s1, s2 = pair
                if not s1.endswith("b") or not s2.endswith("b"):
                    raise ValueError(err_msg)
                try:
                    b1 = int(s1[0:-1])
                    b2 = int(s2[0:-1])
                except ValueError as e:
                    raise ValueError(err_msg) from e
                if b1 > b2:
                    raise ValueError(err_msg)
                mask = 0
                for b in range(b1, b2 + 1):
                    mask |= 1 << b
            else:
                raise ValueError(err_msg)
            masks.append(mask)
            max_mask = max(max_mask, mask)

        for limit, dtype in _MASK_DTYPES:
            if max_mask <= limit:
                return np.array(masks, dtype)

        raise ValueError(err_msg)

    if hasattr(attr_value, "dtype") and hasattr(attr_value, "shape"):
        return attr_value

    if isinstance(attr_value, (list, tuple)):
        return np.array(attr_value)

    raise TypeError(f"attribute {attr_name!r} must be an integer array")


def _sanitize_flag_values(
    flag_var: xr.DataArray,
    flag_values: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    index_tracker = np.arange(len(flag_values))
    fill_value = None
    for d in (flag_var.encoding, flag_var.attrs):
        fill_value = (
            d.get("fill_value", d.get("_FillValue"))
            if fill_value is None
            else fill_value
        )
    if fill_value is not None:
        if np.isnan(fill_value):
            mask = np.logical_not(np.isnan(flag_values))
            index_tracker = index_tracker[mask]
            flag_values = flag_values[mask]
        else:
            mask = flag_values != fill_value
            index_tracker = index_tracker[mask]
            flag_values = flag_values[mask]

    valid_min = flag_var.attrs.get("valid_min")
    valid_max = flag_var.attrs.get("valid_max")
    if valid_min is not None:
        mask = flag_values >= valid_min
        index_tracker = index_tracker[mask]
        flag_values = flag_values[mask]
    if valid_max is not None:
        mask = flag_values <= valid_max
        index_tracker = index_tracker[mask]
        flag_values = flag_values[mask]

    return flag_values, index_tracker



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Hashable, Mapping
from typing import Any, Callable, Optional, Union

import numpy as np
import xarray as xr
from dask import array as da

from xcube.core.gridmapping import GridMapping
from xcube.core.gridmapping.coords import Coords2DGridMapping
from xcube.core.gridmapping.helpers import scale_xy_res_and_size

from .affine import affine_transform_dataset, resample_dataset
from .rectify import rectify_dataset

NDImage = Union[np.ndarray, da.Array]
Aggregator = Callable[[NDImage], NDImage]

# If _SCALE_LIMIT is exceeded, we don't need
# to downscale source image before we can
# rectify it.
_SCALE_LIMIT = 0.95


def resample_in_space(
    source_ds: xr.Dataset,
    /,
    source_gm: Optional[GridMapping] = None,
    target_gm: Optional[GridMapping] = None,
    ref_ds: Optional[xr.Dataset] = None,
    var_configs: Optional[Mapping[Hashable, Mapping[str, Any]]] = None,
    encode_cf: bool = True,
    gm_name: Optional[str] = None,
    rectify_kwargs: Optional[dict] = None,
):
    """
    Resample a dataset *source_ds* in the spatial dimensions.

    If the source grid mapping *source_gm* is not given,
    it is derived from *dataset*:
    ``source_gm = GridMapping.from_dataset(source_ds)``.

    If the target grid mapping *target_gm* is not given,
    it is derived from *source_gm* as
    ``target_gm = source_gm.to_regular()``,
    or if target dataset *ref_ds* is given as
    ``target_gm = GridMapping.from_dataset(ref_ds)``.

    New in 1.6: If *ref_ds* is given, its coordinate
    variables are copied by reference into the returned
    dataset.

    If *source_gm* is almost equal to *target_gm*, this
    function is a no-op and *dataset* is returned unchanged.

    Otherwise, the function computes a spatially
    resampled version of *dataset* and returns it.

    Using *var_configs*, the resampling of individual
    variables can be configured. If given, *var_configs*
    must be a mapping from variable names to configuration
    dictionaries which can have the following properties:

    * ``spline_order`` (int) - The order of spline polynomials
        used for interpolating. It is used for up-sampling only.
        Possible values are 0 to 5.
        Default is 1 (bi-linear) for floating point variables,
        and 0 (= nearest neighbor) for integer and bool variables.
    * ``aggregator`` (str) - An optional aggregating
        function. It is used for down-sampling only.
        Examples are ``numpy.nanmean``, ``numpy.nanmin``,
        ``numpy.nanmax``.
        Default is ``numpy.nanmean`` for floating point variables,
        and None (= nearest neighbor) for integer and bool variables.
    * ``recover_nan`` (bool) - whether a special algorithm
        shall be used that is able to recover values that would
        otherwise yield NaN during resampling.
        Default is False for all variable types since this
        may require considerable CPU resources on top.

    Note that *var_configs* is only used if the resampling involves
    an affine transformation. This is true if the CRS of
    *source_gm* and CRS of *target_gm* are equal and one of two
    cases is given:

    1. *source_gm* is regular.
       In this case the resampling is the affine transformation.
       and the result is returned directly.
    2. *source_gm* is not regular and has a lower resolution
       than *target_cm*.
       In this case *dataset* is down-sampled first using an affine
       transformation. Then the result is rectified.

    In all other cases, no affine transformation is applied and
    the resampling is a direct rectification.

    Args:
        source_ds: The source dataset. Data variables must have 
            dimensions in the following order: optional `time` followed
            by the y-dimension (e.g., `y` or `lat`) followed by the 
            x-dimension (e.g., `x` or `lon`).
        source_gm: The source grid mapping.
        target_gm: The target grid mapping. Must be regular.
        ref_ds: An optional dataset that provides the
            target grid mapping if *target_gm* is not provided.
            If *ref_ds* is given, its coordinate variables are copied
            by reference into the returned dataset.
        var_configs: Optional resampling configurations
            for individual variables.
        encode_cf: Whether to encode the target grid mapping
            into the resampled dataset in a CF-compliant way.
            Defaults to ``True``.
        gm_name: Name for the grid mapping variable.
            Defaults to "crs". Used only if *encode_cf* is ``True``.
        rectify_kwargs: Keyword arguments passed func:`rectify_dataset`
            should a rectification be required.


    Returns: The spatially resampled dataset, or None if the requested
        output area does not intersect with *dataset*.
    """
    if source_gm is None:
        # No source grid mapping given, so do derive it from dataset.
        source_gm = GridMapping.from_dataset(source_ds)

    if target_gm is None:
        # No target grid mapping given, so do derive it
        # from target dataset or source grid mapping.
        if ref_ds is not None:
            target_gm = GridMapping.from_dataset(ref_ds)
        else:
            target_gm = source_gm.to_regular()

    if source_gm.is_close(target_gm):
        # If source and target grid mappings are almost equal.
        # NOTE: Actually we should only return input here if
        # encode_cf == False and gm_name is None and target_ds is None.
        # Otherwise, create a copy and apply encoding and coords copy.
        return source_ds

    # target_gm must be regular
    GridMapping.assert_regular(target_gm, name="target_gm")

    # Are source and target both geographic grid mappings?
    both_geographic = source_gm.crs.is_geographic and target_gm.crs.is_geographic

    if both_geographic or source_gm.crs == target_gm.crs:
        # If CRSes are both geographic or their CRSes are equal:
        if source_gm.is_regular:
            # If also the source is regular, then resampling reduces
            # to an affine transformation.
            return affine_transform_dataset(
                source_ds,
                source_gm=source_gm,
                ref_ds=ref_ds,
                target_gm=target_gm,
                var_configs=var_configs,
                encode_cf=encode_cf,
                gm_name=gm_name,
            )

        # If the source is not regular, we need to rectify it,
        # so the target is regular. Our rectification implementation
        # works only correctly if source pixel size >= target pixel
        # size. Therefore, check if we must downscale source first.
        x_scale = source_gm.x_res / target_gm.x_res
        y_scale = source_gm.y_res / target_gm.y_res
        if x_scale > _SCALE_LIMIT and y_scale > _SCALE_LIMIT:
            # Source pixel size >= target pixel size.
            # We can rectify.
            return rectify_dataset(
                source_ds,
                source_gm=source_gm,
                ref_ds=ref_ds,
                target_gm=target_gm,
                encode_cf=encode_cf,
                gm_name=gm_name,
                **(rectify_kwargs or {}),
            )
        else:
            # Source has higher resolution than target.
            # Downscale first, then rectify
            _, downscaled_size = scale_xy_res_and_size(
                source_gm.xy_res, source_gm.size, (x_scale, y_scale)
            )
            downscaled_dataset = resample_dataset(
                source_ds,
                ((1 / x_scale, 0, 0), (0, 1 / y_scale, 0)),
                size=downscaled_size,
                tile_size=source_gm.tile_size,
                xy_dim_names=source_gm.xy_dim_names,
                var_configs=var_configs,
            )
            downscaled_gm = GridMapping.from_dataset(
                downscaled_dataset,
                tile_size=source_gm.tile_size,
                crs=source_gm.crs,
            )
            return rectify_dataset(
                downscaled_dataset,
                source_gm=downscaled_gm,
                ref_ds=ref_ds,
                target_gm=target_gm,
                encode_cf=encode_cf,
                gm_name=gm_name,
                **(rectify_kwargs or {}),
            )

    # If CRSes are not both geographic and their CRSes are different
    # transform the source_gm so its CRS matches the target CRS:
    transformed_source_gm = source_gm.transform(target_gm.crs, xy_res=target_gm.xy_res)
    if not isinstance(source_gm, Coords2DGridMapping):
        source_ds = source_ds.drop_vars(source_gm.xy_dim_names)
    list_grid_mapping = []
    for var in source_ds.data_vars:
        if "grid_mapping" in source_ds[var].attrs:
            attrs = source_ds[var].attrs
            list_grid_mapping.append(attrs["grid_mapping"])
            del attrs["grid_mapping"]
            source_ds[var] = source_ds[var].assign_attrs(attrs)
    source_ds = source_ds.drop_vars(list_grid_mapping)
    if "crs" in source_ds:
        source_ds = source_ds.drop_vars("crs")
    if "spatial_ref" in source_ds:
        source_ds = source_ds.drop_vars("spatial_ref")
    source_ds = source_ds.copy()
    transformed_x, transformed_y = transformed_source_gm.xy_coords
    attrs = dict(grid_mapping="spatial_ref")
    transformed_x.attrs = attrs
    transformed_y.attrs = attrs
    source_ds = source_ds.assign_coords(
        spatial_ref=xr.DataArray(0, attrs=transformed_source_gm.crs.to_cf()),
        transformed_x=transformed_x,
        transformed_y=transformed_y,
    )
    return resample_in_space(
        source_ds,
        source_gm=transformed_source_gm,
        ref_ds=ref_ds,
        target_gm=target_gm,
        var_configs=var_configs,
        encode_cf=encode_cf,
        gm_name=gm_name,
        rectify_kwargs=rectify_kwargs,
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Sequence
from typing import Any, Callable, Dict, Optional

import xarray as xr

from .abc import MultiLevelDataset
from .lazy import LazyMultiLevelDataset


class CombinedMultiLevelDataset(LazyMultiLevelDataset):
    """A multi-level dataset that is a combination of other
    multi-level datasets.

    Args:
        ml_datasets: The multi-level datasets to be combined. At least
            two must be provided.
        ds_id: Optional dataset identifier.
        combiner_function: An optional function used to combine the
            datasets, for example ``xarray.merge``. If given, it
            receives a list of datasets (``xarray.Dataset`` instances)
            and *combiner_params* as keyword arguments. If not given or
            ``None`` is passed, a copy of the first dataset is made,
            which is then subsequently updated by the remaining datasets
            using ``xarray.Dataset.update()``.
        combiner_params: Parameters to the *combiner_function* passed as
            keyword arguments.
    """

    def __init__(
        self,
        ml_datasets: Sequence[MultiLevelDataset],
        ds_id: Optional[str] = None,
        combiner_function: Optional[Callable] = None,
        combiner_params: Optional[dict[str, Any]] = None,
    ):
        if not ml_datasets or len(ml_datasets) < 2:
            raise ValueError("ml_datasets must have at least two elements")
        super().__init__(ds_id=ds_id, parameters=combiner_params)
        self._ml_datasets = ml_datasets
        self._combiner_function = combiner_function

    def _get_num_levels_lazily(self) -> int:
        return self._ml_datasets[0].num_levels

    def _get_dataset_lazily(
        self, index: int, combiner_params: dict[str, Any]
    ) -> xr.Dataset:
        datasets = [ml_dataset.get_dataset(index) for ml_dataset in self._ml_datasets]
        if self._combiner_function is None:
            combined_dataset = datasets[0].copy()
            for dataset in datasets[1:]:
                combined_dataset.update(dataset)
            return combined_dataset
        else:
            return self._combiner_function(datasets, **combiner_params)

    def close(self):
        for ml_dataset in self._ml_datasets:
            ml_dataset.close()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import itertools
from collections.abc import Iterable, Iterator
from typing import Dict, Tuple

import numpy as np
import xarray as xr

from xcube.core.update import update_dataset_chunk_encoding


def chunk_dataset(
    dataset: xr.Dataset,
    chunk_sizes: dict[str, int] = None,
    format_name: str = None,
    data_vars_only: bool = False,
) -> xr.Dataset:
    """Chunk *dataset* using *chunk_sizes* and optionally
    update encodings for given *format_name*.

    Args:
        dataset: input dataset
        chunk_sizes: mapping from dimension name to new chunk size
        format_name: optional format, e.g. "zarr" or "netcdf4"
        data_vars_only: only chunk data variables, not coordinates

    Returns:
        the (re)chunked dataset
    """

    if data_vars_only:
        for variable in dataset.data_vars:
            dataset[variable] = dataset[variable].chunk(chunk_sizes)
    else:
        dataset = dataset.chunk(chunks=chunk_sizes)
    if format_name:
        dataset = update_dataset_chunk_encoding(
            dataset,
            chunk_sizes=chunk_sizes,
            format_name=format_name,
            data_vars_only=data_vars_only,
        )
    return dataset


def get_empty_dataset_chunks(
    dataset: xr.Dataset,
) -> Iterator[tuple[str, Iterator[tuple[int, ...]]]]:
    """Identify empty dataset chunks and return their indices.

    Args:
        dataset: The dataset.

    Returns:
        An iterator that provides a stream of (variable name, block
        indices tuple) tuples.
    """
    return (
        (str(var_name), get_empty_var_chunks(dataset[var_name]))
        for var_name in dataset.data_vars
    )


def get_empty_var_chunks(var: xr.DataArray) -> Iterator[tuple[int, ...]]:
    """Identify empty variable chunks and return their indices.

    Args:
        var: The variable.

    Returns:
        A list of block indices.
    """
    chunks = var.chunks
    if chunks is None:
        return None

    for chunk_index, chunk_slice in compute_chunk_slices(chunks):
        data_index = tuple(slice(start, end) for start, end in chunk_slice)
        data = var[data_index]
        if np.all(np.isnan(data)):
            # print(f'empty: {var.name}/{".".join(map(str, chunk_index))}')
            yield chunk_index


def compute_chunk_slices(chunks: tuple[tuple[int, ...], ...]) -> Iterable:
    chunk_indices = []
    for c in chunks:
        chunk_indices.append(tuple(i for i in range(len(c))))

    chunk_slices = []
    for c in chunks:
        x = []
        o = 0
        for s in c:
            x.append((o, o + s))
            o += s
        chunk_slices.append(tuple(x))

    return zip(itertools.product(*chunk_indices), itertools.product(*chunk_slices))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


import importlib
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Dict, List, Optional

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

from xcube.util.ipython import register_json_formatter

Component = Any
ComponentLoader = Callable[["Extension"], Component]
ComponentTransform = Callable[[Component, "Extension"], Component]
ExtensionPredicate = Callable[["Extension"], bool]


class Extension:
    """An extension that provides a component of any type.

    Extensions are registered in a :class:`ExtensionRegistry`.

    Extension objects are not meant to be instantiated directly. Instead,
    :meth:`ExtensionRegistry#add_extension` is used to register extensions.

    Args:
        point: extension point identifier
        name: extension name
        component: extension component
        loader: extension component loader function
        metadata: extension metadata
    """

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        point: str,
        name: str,
        component: Component = None,
        loader: ComponentLoader = None,
        **metadata,
    ):
        if point is None:
            raise ValueError(f"point must be given")
        if name is None:
            raise ValueError(f"name must be given")
        if (loader is not None and component is not None) or (
            loader is None and component is None
        ):
            raise ValueError(f"either component or loader must be given")
        if loader is not None and not callable(loader):
            raise ValueError(f"loader must be callable")

        self._component = component
        self._loader = loader
        self._point = point
        self._name = name
        self._metadata = metadata
        self._deleted = False

    @property
    def is_lazy(self) -> bool:
        """Whether this is a lazy extension that uses a loader."""
        return self._loader is not None

    @property
    def component(self) -> Component:
        """Extension component."""
        if self._component is None and self._loader is not None:
            self._component = self._loader(self)
        return self._component

    @property
    def point(self) -> str:
        """Extension point identifier."""
        return self._point

    @property
    def name(self) -> str:
        """Extension name."""
        return self._name

    @property
    def metadata(self) -> dict[str, Any]:
        """Extension metadata."""
        return dict(self._metadata)

    def to_dict(self) -> dict[str, Any]:
        """Get a JSON-serializable dictionary representation of this extension."""

        # Note: we avoid loading the component!
        if self._component is not None:
            if hasattr(self._component, "to_dict") and callable(
                getattr(self._component, "to_dict")
            ):
                component = self._component.to_dict()
            else:
                component = repr(self._component)
        else:
            component = "<not loaded yet>"
        d = dict(
            name=self.name,
            **self.metadata,
            point=self.point,
            component=component,
        )
        return d


register_json_formatter(Extension)


# noinspection PyShadowingBuiltins
class ExtensionRegistry:
    """A registry of extensions.
    Typically used by plugins to register extensions.
    """

    def __init__(self):
        self._extension_points = {}

    def has_extension(self, point: str, name: str) -> bool:
        """Test if an extension with given *point* and *name* is registered.

        Args:
            point: extension point identifier
            name: extension name

        Returns:
            True, if extension exists
        """
        return point in self._extension_points and name in self._extension_points[point]

    def get_extension(self, point: str, name: str) -> Optional[Extension]:
        """Get registered extension for given *point* and *name*.

        Args:
            point: extension point identifier
            name: extension name

        Returns:
            the extension or None, if no such exists
        """
        if point not in self._extension_points:
            return None
        return self._extension_points[point].get(name)

    def get_component(self, point: str, name: str) -> Any:
        """Get extension component for given *point* and *name*.
        Raises a ValueError if no such extension exists.

        Args:
            point: extension point identifier
            name: extension name

        Returns:
            extension component
        """
        extension = self.get_extension(point, name)
        if extension is None:
            raise ValueError(
                f"extension {name!r} not found for extension point {point!r}"
            )
        return extension.component

    def find_extensions(
        self, point: str, predicate: ExtensionPredicate = None
    ) -> list[Extension]:
        """Find extensions for *point* and optional filter function *predicate*.

        The filter function is called with an extension and should return
        a truth value to indicate a match or mismatch.

        Args:
            point: extension point identifier
            predicate: optional filter function

        Returns:
            list of matching extensions
        """
        if point not in self._extension_points:
            return []
        point_extensions = self._extension_points[point]
        if predicate is None:
            return list(point_extensions.values())
        return [
            extension for extension in point_extensions.values() if predicate(extension)
        ]

    def find_components(
        self, point: str, predicate: ExtensionPredicate = None
    ) -> list[Component]:
        """Find extension components for *point* and optional filter function *predicate*.

        The filter function is called with an extension and should return
        a truth value to indicate a match or mismatch.

        Args:
            point: extension point identifier
            predicate: optional filter function

        Returns:
            list of matching extension components
        """
        return [
            extension.component
            for extension in self.find_extensions(point, predicate=predicate)
        ]

    def add_extension(
        self,
        point: str,
        name: str,
        component: Component = None,
        loader: ComponentLoader = None,
        **metadata,
    ) -> Extension:
        """Register an extension *component* or an extension component *loader* for
        the given extension *point*, *name*, and additional *metadata*.

        Either *component* or *loader* must be specified, but not both.

        A given *loader* must be a callable with one positional argument *extension* of type :class:`Extension`
        and is expected to return the actual extension component, which may be of any type.
        The *loader* will only be called once and only when the actual extension component
        is requested for the first time. Consider using the function :func:`import_component` to create a
        loader that lazily imports a component from a module and optionally executes it.

        Args:
            point: extension point identifier
            name: extension name
            component: extension component
            loader: extension component loader function
            **metadata: extension metadata

        Returns:
            a registered extension
        """
        extension = Extension(
            point, name, component=component, loader=loader, **metadata
        )
        if point in self._extension_points:
            self._extension_points[point][name] = extension
        else:
            self._extension_points[point] = {name: extension}
        return extension

    def remove_extension(self, point: str, name: str):
        """Remove registered extension *name* from given *point*.

        Args:
            point: extension point identifier
            name: extension name
        """
        point_extensions = self._extension_points[point]
        del point_extensions[name]

    def to_dict(self):
        """Get a JSON-serializable dictionary representation of this extension registry."""
        return {
            k: {ek: ev.to_dict() for ek, ev in v.items()}
            for k, v in self._extension_points.items()
        }


register_json_formatter(ExtensionRegistry)

_EXTENSION_REGISTRY_SINGLETON = ExtensionRegistry()


def get_extension_registry() -> ExtensionRegistry:
    """Return the extension registry singleton."""
    return _EXTENSION_REGISTRY_SINGLETON


def import_component(
    spec: str,
    transform: ComponentTransform = None,
    call: bool = False,
    call_args: Sequence[Any] = None,
    call_kwargs: Mapping[str, Any] = None,
) -> ComponentLoader:
    """Return a component loader that imports a module or module component from *spec*.
    To import a module, *spec* should be the fully qualified module name. To import a
    component, *spec* must also append the component name to the fully qualified module name
    separated by a color (":") character.

    An optional *transform* callable my be used to transform the imported component. If given,
    a new component is computed::

        component = transform(component, extension)

    If the *call* flag is set, the component is expected to be a callable which will be called
    using the given *call_args* and *call_kwargs* to produce a new component::

        component = component(*call_kwargs, **call_kwargs)

    Finally, the component is returned.

    Args:
        spec: String of the form "module_path" or
            "module_path:component_name"
        transform: callable that takes two positional arguments, the
            imported component and the extension of type
            :class:`Extension`
        call: Whether to finally call the component with given
            *call_args* and *call_kwargs*
        call_args: arguments passed to a callable component if *call*
            flag is set
        call_kwargs: keyword arguments passed to callable component if
            *call* flag is set

    Returns:
        a component loader
    """

    # noinspection PyUnusedLocal
    def _load(extension: Extension):
        nonlocal spec
        component = _import_component(spec, force_component=call)
        if transform is not None:
            component = transform(component, extension)
        if call or call_args or call_kwargs:
            component = component(*(call_args or []), **(call_kwargs or {}))
        return component

    return _load


def _import_component(component_spec: str, force_component: bool = False):
    """Import a module or module component from *spec*.

    Args:
        component_spec: String of the form "module_name" or
            "module_name:component_name" where module_name must an
            absolute, fully qualified path to a module.
        force_component: If True, *spec* must specify a component name

    Returns:
        the imported module or module component
    """
    if ":" in component_spec:
        module_name, component_name = component_spec.split(":", maxsplit=1)
    else:
        module_name, component_name = component_spec, None

    if force_component and component_name is None:
        raise ValueError("illegal spec, must specify a component")
    if module_name == "":
        raise ValueError("illegal spec, missing module path")
    if component_name == "":
        raise ValueError("illegal spec, missing component name")

    module = importlib.import_module(module_name)
    return getattr(module, component_name) if component_name else module



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiContext, Context

from ..datasets.context import DatasetsContext


class StatisticsContext(ApiContext):
    def __init__(self, server_ctx: Context):
        super().__init__(server_ctx)
        self._datasets_ctx = server_ctx.get_api_ctx("datasets")

    @property
    def datasets_ctx(self) -> DatasetsContext:
        return self._datasets_ctx



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import itertools
import json
from collections.abc import MutableMapping, Sequence
from typing import Any, Callable, Dict, Tuple, Union

import numpy as np

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

import collections.abc
from collections.abc import Iterable, Iterator, KeysView
from logging import Logger
from typing import Optional

from deprecated import deprecated

from xcube.constants import LOG, LOG_LEVEL_TRACE
from xcube.core.zarrstore import LoggingZarrStore
from xcube.util.assertions import assert_instance

GetChunk = Callable[["ChunkStore", str, tuple[int, ...]], bytes]


# Note, we cannot remove this deprecated code as long as
# xcube.core.compute.compute_dataset() is using it.
@deprecated(
    reason="This class shall no longer be used."
    " If similar functionality is needed,"
    " use xcube.core.zarrstore.GenericZarrStore",
    version="0.12.1",
)
class ChunkStore(MutableMapping):
    """A Zarr Store that generates datasets by allowing data variables to
    fetch or compute their chunks by a user-defined function *get_chunk*.
    Implements the standard Python ``MutableMapping`` interface.

    This is how the *get_chunk* function is called:::

        data = get_chunk(chunk_store, var_name, chunk_indexes)

    where ``chunk_store`` is this store, ``var_name`` is the name of
    the variable for which data is fetched, and ``chunk_indexes`` is
    a tuple of zero-based, integer chunk indexes. The result must be
    a Python *bytes* object.

    Args:
        dims: Dimension names of all data variables,
            e.g. ('time', 'lat', 'lon').
        shape: Shape of all data variables according to *dims*,
            e.g. (512, 720, 1480).
        chunks: Chunk sizes of all data variables according to *dims*,
            e.g. (128, 180, 180).
        attrs: Global dataset attributes.
        get_chunk: Default chunk fetching/computing function.
        trace_store_calls: Whether to log calls
            into the ``MutableMapping`` interface.
    """

    def __init__(
        self,
        dims: Sequence[str],
        shape: Sequence[int],
        chunks: Sequence[int],
        attrs: dict[str, Any] = None,
        get_chunk: GetChunk = None,
        trace_store_calls: bool = False,
    ):
        self._ndim = len(dims)
        self._dims = tuple(dims)
        self._shape = tuple(shape)
        self._chunks = tuple(chunks)
        self._get_chunk = get_chunk
        self._trace_store_calls = trace_store_calls

        # setup Virtual File System (vfs)
        self._vfs = {
            ".zgroup": _dict_to_bytes(dict(zarr_format=2)),
            ".zattrs": _dict_to_bytes(attrs or dict()),
        }

    @property
    def ndim(self) -> int:
        return self._ndim

    @property
    def dims(self) -> tuple[str, ...]:
        return self._dims

    @property
    def shape(self) -> tuple[int, ...]:
        return self._shape

    @property
    def chunks(self) -> tuple[int, ...]:
        return self._chunks

    def add_array(self, name: str, array: np.ndarray, attrs: dict):
        shape = list(map(int, array.shape))
        dtype = str(array.dtype.str)
        array_metadata = {
            "zarr_format": 2,
            "chunks": shape,
            "shape": shape,
            "dtype": dtype,
            "fill_value": None,
            "compressor": None,
            "filters": None,
            "order": "C",
        }
        self._vfs[name] = _str_to_bytes("")
        self._vfs[name + "/.zarray"] = _dict_to_bytes(array_metadata)
        self._vfs[name + "/.zattrs"] = _dict_to_bytes(attrs)
        self._vfs[name + "/" + (".".join(["0"] * array.ndim))] = bytes(array)

    def add_lazy_array(
        self,
        name: str,
        dtype: str,
        fill_value: Union[int, float] = None,
        compressor: dict[str, Any] = None,
        filters=None,
        order: str = "C",
        attrs: dict[str, Any] = None,
        get_chunk: GetChunk = None,
    ):
        get_chunk = get_chunk or self._get_chunk
        if get_chunk is None:
            raise ValueError("get_chunk must be given as there is no default")

        array_metadata = dict(
            zarr_format=2,
            shape=self._shape,
            chunks=self._chunks,
            compressor=compressor,
            dtype=dtype,
            fill_value=fill_value,
            filters=filters,
            order=order,
        )

        self._vfs[name] = _str_to_bytes("")
        self._vfs[name + "/.zarray"] = _dict_to_bytes(array_metadata)
        self._vfs[name + "/.zattrs"] = _dict_to_bytes(
            dict(_ARRAY_DIMENSIONS=self._dims, **(attrs or dict()))
        )

        nums = np.array(self._shape) // np.array(self._chunks)
        indexes = itertools.product(*tuple(map(range, map(int, nums))))
        for index in indexes:
            filename = ".".join(map(str, index))
            # noinspection PyTypeChecker
            self._vfs[name + "/" + filename] = name, index, get_chunk

    @property
    def _class_name(self):
        return self.__module__ + "." + self.__class__.__name__

    ##########################################################################
    # Zarr Store (MutableMapping) implementation
    ##########################################################################

    def keys(self) -> KeysView[str]:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.keys()")
        return self._vfs.keys()

    def listdir(self, key: str) -> Iterable[str]:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.listdir(key={key!r})")
        if key == "":
            return (k for k in self._vfs.keys() if "/" not in k)
        else:
            prefix = key + "/"
            start = len(prefix)
            return (
                k
                for k in self._vfs.keys()
                if k.startswith(prefix) and k.find("/", start) == -1
            )

    def getsize(self, key: str) -> int:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.getsize(key={key!r})")
        return len(self._vfs[key])

    def __iter__(self) -> Iterator[str]:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.__iter__()")
        return iter(self._vfs.keys())

    def __len__(self) -> int:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.__len__()")
        return len(self._vfs.keys())

    def __contains__(self, key) -> bool:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.__contains__(key={key!r})")
        return key in self._vfs

    def __getitem__(self, key: str) -> bytes:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.__getitem__(key={key!r})")
        value = self._vfs[key]
        if isinstance(value, tuple):
            name, index, get_chunk = value
            return get_chunk(self, name, index)
        return value

    def __setitem__(self, key: str, value: bytes) -> None:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.__setitem__(key={key!r}, value={value!r})")
        raise TypeError(f"{self._class_name} is read-only")

    def __delitem__(self, key: str) -> None:
        if self._trace_store_calls:
            _trace(f"{self._class_name}.__delitem__(key={key!r})")
        raise TypeError(f"{self._class_name} is read-only")


def _trace(msg: str):
    LOG.log(LOG_LEVEL_TRACE, msg)


def _dict_to_bytes(d: dict):
    return _str_to_bytes(json.dumps(d, indent=2))


def _str_to_bytes(s: str):
    return bytes(s, encoding="utf-8")


@deprecated(
    reason="This class has been moved,"
    " use xcube.core.zarrstore.LoggingZarrStore"
    " instead.",
    version="0.12.1",
)
class LoggingStore(LoggingZarrStore):
    """A Zarr Store that logs all method calls on another store *other*
    including execution time.
    """

    @classmethod
    def new(
        cls,
        other: collections.abc.MutableMapping,
        logger: Logger = LOG,
        name: Optional[str] = None,
    ):
        assert_instance(other, collections.abc.MutableMapping)
        return cls(other, logger=logger, name=name)


@deprecated(
    reason="This class has been moved,"
    " use xcube.core.zarrstore.LoggingZarrStore"
    " instead.",
    version="0.12.1",
)
class MutableLoggingStore(LoggingStore):
    """Mutable version of :class:`LoggingStore`."""



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import sys
from typing import Dict, List

DEFAULT_DEPENDENCY_NAMES = [
    "affine",
    "click",
    "cmocean",
    "dask",
    "dask - image",
    "distributed",
    "fiona",
    "fsspec",
    "gdal",
    "geopandas",
    "jdcal",
    "jsonschema",
    "matplotlib - base",
    "netcdf4",
    "numba",
    "numpy",
    "pandas",
    "pillow",
    "pyjwt",
    "pyproj",
    "pyyaml",
    "rasterio",
    "requests",
    "requests-oauthlib",
    "s3fs",
    "scipy",
    "setuptools",
    "shapely",
    "tornado",
    "urllib3",
    "xarray",
    "zarr",
]

_XCUBE_VERSIONS = None


def get_xcube_versions() -> dict[str, str]:
    """Get a mapping from xcube package names to package versions.

    Returns:
        A mapping of the package names to package versions
    The result computed from the evaluating the expression.
    """
    from .plugin import get_plugins

    global _XCUBE_VERSIONS
    if _XCUBE_VERSIONS is None:
        plugin_names = [f"{plugin}.version" for plugin in list(get_plugins().keys())]

        _XCUBE_VERSIONS = get_versions(DEFAULT_DEPENDENCY_NAMES, plugin_names)
    return _XCUBE_VERSIONS


def get_versions(
    dependency_names: list[str], plugin_names: list[str]
) -> dict[str, str]:
    """Get a mapping from package names to package versions.
    The input is divided into names of packages that are external dependencies
    and into names of xcube plugins.

    Args:
        dependency_names: A list of names of packages of which the
            versions shall be found
        plugin_names: A list of names of xcube plugins of which the
            versions shall be found

    Returns:
        A mapping of the package names to package versions
    The result computed from the evaluating the expression.
    """
    # Idea borrowed from xarray.print_versions
    import importlib

    def _maybe_add_dot_version(name: str):
        if not name.endswith(".version"):
            return f"{name}.version"
        return name

    dependencies = [
        (_maybe_add_dot_version(plugin_name), lambda mod: mod.version)
        for plugin_name in plugin_names
    ]

    def _find_module_version(mod):
        if hasattr(mod, "__version__"):
            return mod.__version__
        elif hasattr(mod, "version"):
            return mod.version
        else:
            return "unknown"

    dependencies += [
        (dependency_name, _find_module_version) for dependency_name in dependency_names
    ]

    dependencies_dict = {}
    for module_name, module_version in dependencies:
        module_key = module_name.split(".")[0]
        # noinspection PyBroadException
        try:
            if module_name in sys.modules:
                module = sys.modules[module_name]
            else:
                module = importlib.import_module(module_name)
        except BaseException:
            pass
        else:
            # noinspection PyBroadException
            try:
                dependencies_dict[module_key.split(".version")[0]] = module_version(
                    module
                )
            except BaseException as e:
                dependencies_dict[module_key.split(".version")[0]] = "installed"

    return dependencies_dict



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


class VarExprError(ValueError):
    """Exception raised by code in the ``xcube.code.varexpr` module."""
